from typing import List, Union, Dict
import os, pickle, sys
import pathlib
from collections import deque
import collections
import pandas as pd
import numpy as np
from io import StringIO
import ast
import multiprocessing
import torch
from torchtext import data
import torchtext


from src.data.basis import BasisDataSet

"""
Loads data from:
archiveII (e2efold)
RNAStrAlign (e2efold)
RNA_STAND (http://www.rnasoft.ca/strand/)
bpRNA (http://bprna.cgrb.oregonstate.edu/download.php#bpRNA)

"""
RNA_SS_data = collections.namedtuple('RNA_SS_data', 'seq ss_label length name pairs')

class RNAct(BasisDataSet):

    def __init__(self, data_dir, max_len, min_len, source, include_N, design=False, partial=False):

        super().__init__(data_dir=data_dir)

        self.rng = np.random.RandomState(seed=100)

        self.samples = {"train": [], "valid": [], "test": []}
        self.split = {"train": 0.8, "valid": 0.1, "test": 0.1}

        dump = True

        self.max_len = max_len
        self.min_len = min_len
        self.include_N = include_N
        self.design  = design

        if not isinstance(partial, int):
            partial = 10
        self.partial  = partial


        self.data_dir = self.data_dir


        config_str = f"{max_len}-{min_len}-{include_N}-{''.join(source) }-design-{design}-partial-{partial}"
        data_file = f".data/RNAct_hash-{config_str}.plk"
        if os.path.isfile(data_file):
            with open(data_file, "rb") as f:
                load_out = pickle.load(f)

                if self.partial:
                    train_list, valid_list, test_list, self.SRC, self.TRG, self.POS1ID, self.POS2ID, self.PK, self.MASK = load_out
                    fields = [("src", self.SRC), ("trg", self.TRG), ("pos1id", self.POS1ID), ("pos2id", self.POS2ID), ("pk", self.PK), ("mask", self.MASK)]
                else:
                    train_list, valid_list, test_list, self.SRC, self.TRG, self.POS1ID, self.POS2ID, self.PK = load_out
                    fields = [("src", self.SRC), ("trg", self.TRG), ("pos1id", self.POS1ID),("pos2id", self.POS2ID),("pk", self.PK)]

                self.train = torchtext.data.Dataset(examples=train_list, fields=fields)
                self.valid = torchtext.data.Dataset(examples=valid_list, fields=fields)
                self.test = torchtext.data.Dataset(examples=test_list, fields=fields)
            print(f"RNA ct file: reload existing data dump {data_file}")
        else:

            file_list = self.get_file_list(source)
            sample_df = self.process_ct_files(file_list)

            self.create_datasets(sample_df)

            self.SRC.build_vocab(self.train.src, min_freq=2)
            self.TRG.build_vocab(self.train.trg, min_freq=2)

            if dump:
                with open(data_file, "wb") as f:
                    if self.partial:
                        work_load = [self.train.examples, self.valid.examples, self.test.examples, self.SRC, self.TRG, self.POS1ID, self.POS2ID, self.PK, self.MASK]
                    else:
                        work_load = [self.train.examples, self.valid.examples, self.test.examples, self.SRC, self.TRG, self.POS1ID, self.POS2ID, self.PK ]
                    pickle.dump(work_load, f)
                print("# RNAct dump successful")


    def get_file_list(self, source):
        file_list = []
        if "RNAStrAlign" in source:
            rna_types = ['tmRNA', 'tRNA', 'telomerase', 'RNaseP', 'SRP', '16S_rRNA', '5S_rRNA', 'group_I_intron']
            for rna_type in rna_types:
                file_list.extend(self.find_ct_files(dir=self.data_dir / "RNAStrAlign" / (rna_type + '_database'), rna_type=rna_type))
        if "archiveII" in source:
            file_list.extend(self.find_ct_files(dir=self.data_dir / "archiveII"))
        if "RNA_Stand" in source:
            file_list.extend(self.find_ct_files(dir=self.data_dir / "RNA_STRAND_data"))
        if "bpRNA" in source:
            file_list.extend(self.find_ct_files(dir=self.data_dir / "bpRNA_dataset/ctFiles", type_segment=1))
        return file_list


    def find_ct_files(self, dir, rna_type=None, type_segment=0):
        file_list = []
        for r, d, f in os.walk(dir):
            for file in f:
                if file.endswith(".ct"):
                    if rna_type:
                        file_list.append({"dir": os.path.join(r, file), "file": file, "type": rna_type, "data_set": dir.parts[-2]})
                    else:
                        file_list.append({"dir": os.path.join(r, file), "file": file, "type": file.split("_")[type_segment], "data_set": dir.parts[-1]})
        return file_list


    def _read_ct_file(self, file_dir:str):

        try:
            lines = "".join([line for line in open(file_dir)
                             if not line.startswith("#")])

            sample = pd.read_csv(StringIO(lines), sep='\s+', skiprows=1, header=None)
        except pd.errors.ParserError:
            print("ParserError", file_dir)
            return False

        sequence = sample.loc[:, 1].to_numpy().tolist()
        if  not self.min_len <= len(sequence) <= self.max_len:
            return False


        pos1id_list = list(sample.loc[:, 0].values)
        pos2id_list = list(sample.loc[:, 4].values)
        pairs_list = list(zip(pos1id_list, pos2id_list))
        pairs_list = list(filter(lambda x: x[1] > 0, pairs_list))
        # pairs_list = list(filter(lambda x: x[1] > x[0], pairs_list))


        if len(pairs_list) < 2:             # TODO include no pair sequcneces
            print("len(paris) < 2", file_dir)
            return False

        if len(pairs_list) % 2 == 1:
            print("len(paris) % 2 = 1", file_dir)
            return False

        norm_numb = pos1id_list[0]
        pairs = np.array(pairs_list) - norm_numb

        if len(sequence) <= np.max(pairs):
            print("len(sequence) <= max(posID)", file_dir)
            return False

        pairs = list(pairs)
        structure, pk_pairs = self.pairs_to_dot_point(pairs, len(sequence))

        pk_pairs = np.asarray(pk_pairs)
        if len(pk_pairs.shape) <= 1:
            print("len(pk_pairs.shape) <= 1", file_dir)
            return False
        pk_pairs = pk_pairs[np.lexsort((pk_pairs[:, 0], pk_pairs[:, 2]))]
        pos1id = pk_pairs[:, 0].tolist()
        pos2id = pk_pairs[:, 1].tolist()
        pseudoknot = pk_pairs[:, 2].tolist()

        return {"structure": structure, "sequence":sequence, "pos1id":pos1id, "pos2id":pos2id, "pk": pseudoknot}


    @staticmethod # also used in RNAspotrna
    def pairs_to_dot_point(pairs, length, inverse=False):
        structure = ['.'] * length
        pk_pairs = []

        if len(pairs) == 0:
            return structure, pk_pairs

        pairs = np.stack(pairs, axis=0)
        if inverse:
            pairs = pairs[pairs[:, 0].argsort()[::-1]]
        else:
            pairs = pairs[pairs[:, 0].argsort()]
        pairs = [pairs[i, :] for i in range(pairs.shape[0])]
        pair_buffer = [deque(), deque(), deque(), deque()]

        for i, j in pairs:

            in_level = -1
            for level, buffer in enumerate(pair_buffer):
                if (i, j) in buffer:
                    in_level = level
                    break
            if in_level == -1:
                pair_buffer[0].append((j, i))
            else:
                while True:
                    if (i, j) == pair_buffer[in_level][-1]:
                        pair_buffer[in_level].pop()
                        if inverse:
                            structure[j] = f"){in_level}"
                            structure[i] = f"({in_level}"
                            pk_pairs.append([i, j, in_level])
                        else:
                            structure[j] = f"({in_level}"
                            structure[i] = f"){in_level}"
                            pk_pairs.append([j, i, in_level])
                        break
                    else:
                        pair_buffer[min(in_level + 1, len(pair_buffer) - 1)].insert(0, pair_buffer[in_level].pop())
        return structure, pk_pairs





    def process_ct_files(self, file_list):

        print("file list len: ", len(file_list))

        sample_list = []
        for file_dict in file_list:
            sample_dict = self._read_ct_file(file_dict['dir'])
            if sample_dict:
                sample_dict.update(file_dict)
                sample_list.append(sample_dict)

        sample_df = pd.DataFrame(sample_list)

        sample_df = sample_df[sample_df['sequence'].apply(lambda x: self.min_len <= len(x) <= self.max_len)]
        print("samples after length drop",  sample_df.shape[0])

        if self.include_N:
            bases = ["A", "C", "G", "U", "N"]
        else:
            bases = ["A", "C", "G", "U"]
        sample_df = sample_df[sample_df['sequence'].apply(lambda x: all([s in bases for s in x])   )]
        print("samples after wrong bases drop", sample_df.shape[0])

        sample_df['seq_struc_str'] = sample_df['sequence'].apply(lambda x: ''.join(x)) + sample_df['structure'].apply(lambda x: ''.join(x))
        # drop equal sequences
        sample_df = sample_df.loc[~sample_df['seq_struc_str'].duplicated()]
        sample_df = sample_df.drop(columns=['seq_struc_str', 'dir'])
        print("samples after drop duplicated", sample_df.shape[0])

        sample_df = sample_df.reset_index()

        return sample_df

    def save_to_file(self, sample_df:pd.DataFrame):
        for idx in range(sample_df.shape[0]):
            with open(f'./rnajoined/rnajoined_{idx//500:06}.fasta', 'a') as f:
                seq = ''.join(sample_df['sequence'][idx])
                name = f">{sample_df['data_set'][idx]}_{sample_df['type'][idx]}_{idx:06}"
                f.write(name + "\n")
                f.write(seq + "\n")


    def split_dataframe(self, samples_df:pd.DataFrame ):

        train_df = pd.DataFrame()
        valid_df = pd.DataFrame()
        test_df = pd.DataFrame()

        rna_types = samples_df['type'].unique().tolist()
        for rtype in rna_types:
            sub_df = samples_df.loc[samples_df['type'] == rtype]

            rand_idx = self.rng.permutation(sub_df.shape[0])

            train_idx = int(sub_df.shape[0] * self.split['train'])
            valid_idx = int(sub_df.shape[0] * self.split['valid']) + train_idx

            sub_train_df = sub_df.iloc[rand_idx[:train_idx]]
            sub_valid_df = sub_df.iloc[rand_idx[train_idx:valid_idx]]
            sub_test_df = sub_df.iloc[rand_idx[valid_idx:]]

            train_df = pd.concat([train_df, sub_train_df]) #train_df.append(sub_train_df, ignore_index=True)
            valid_df = pd.concat([valid_df, sub_valid_df]) #valid_df.append(sub_valid_df, ignore_index=True)
            test_df = pd.concat([test_df, sub_test_df]) #test_df.append(sub_test_df, ignore_index=True )

            # print(f"family: {rtype} rand_idx", rand_idx.shape)
            # print(f"family: {rtype} train_idx", train_idx)
            print(f"family: {rtype} sub_train_df", sub_train_df.shape)
            print(f"family: {rtype} train_df", train_df.shape)

        train_df = train_df.reset_index()
        valid_df = valid_df.reset_index()
        test_df = test_df.reset_index()

        return train_df,  valid_df,  test_df


    def create_sample_list(self, sample_df):

        sample_list = []
        for idx, sample in sample_df.iterrows():
            sample_list.append({"sequence": sample["sequence"], "structure": sample["structure"], "pos1id": sample["pos1id"]
                                   , "pos2id": sample["pos2id"], "pk": sample["pk"]})

        return sample_list




    def create_datasets(self, sample_df):

        train_df,  valid_df,  test_df = self.split_dataframe(sample_df)

        print("train_df", train_df.shape)
        print("valid_df", valid_df.shape)
        print("test_df ", test_df.shape)

        samples = {}
        train_sample_list = self.create_sample_list(train_df)
        train_sample_list = self.rng.permutation(train_sample_list)
        samples['train'] = train_sample_list
        samples['valid'] = self.create_sample_list(valid_df)
        samples['test'] = self.create_sample_list(test_df)

        self.SRC = data.Field(pad_token=self.blank_word, unk_token=None, batch_first=True, include_lengths=True)
        self.TRG = data.Field(pad_token=self.blank_word, unk_token=None, batch_first=True, include_lengths=True)
        self.POS1ID = data.Field(pad_token=-1, batch_first=True, include_lengths=True, use_vocab=False)
        self.POS2ID = data.Field(pad_token=-1, batch_first=True, include_lengths=True, use_vocab=False)
        self.PK = data.Field(pad_token=-1, batch_first=True, include_lengths=True, use_vocab=False)

        fields = {"src": ("src", self.SRC), "trg": ("trg", self.TRG), "pos1id": ("pos1id", self.POS1ID),
                  "pos2id": ("pos2id", self.POS2ID),"pk": ("pk", self.PK)}

        if self.partial:
            self.MASK = data.Field(pad_token=-1, batch_first=True, include_lengths=True, use_vocab=False)
            fields = {"src": ("src", self.SRC), "trg": ("trg", self.TRG), "pos1id": ("pos1id", self.POS1ID),
                      "pos2id": ("pos2id", self.POS2ID), "pk": ("pk", self.PK), "mask": ("mask", self.MASK)}



        for set, sample_list in samples.items():
            example_list = []
            for sample in sample_list:

                if self.partial:
                    structure = sample.pop("structure")
                    sequence = sample.pop("sequence")

                    if set == "train":
                        for i in range(self.partial):
                            src, trg, mask = self.make_partial(sequence=sequence, structure=structure)
                            sample["src"] = src
                            sample["trg"] = trg
                            sample["mask"] = mask
                            example_list.append(torchtext.data.Example().fromdict(sample, fields=fields))
                    else:
                        # src, trg, mask = self.make_partial(sequence=sequence, structure=structure)
                        # sample["src"] = src
                        # sample["trg"] = trg
                        # sample["mask"] = mask

                        sample["src"] = structure
                        sample["trg"] = sequence + [self.blank_word] * len(sample["src"])
                        sample["mask"] = [1]*len(sample["src"]) + [0]*len(sample["src"])
                        example_list.append(torchtext.data.Example().fromdict(sample, fields=fields))


                elif self.design:
                    sample["src"] = sample.pop("structure")
                    sample["trg"] = sample.pop("sequence")
                    example_list.append(torchtext.data.Example().fromdict(sample, fields=fields))
                else:
                    sample["src"] = sample.pop("sequence")
                    sample["trg"] = sample.pop("structure")
                    example_list.append(torchtext.data.Example().fromdict(sample, fields=fields))



            setattr(self, set, torchtext.data.Dataset(example_list, fields=list(fields.values())))


    def make_partial(self, sequence, structure):
        trg_structure = []
        trg_sequence = []
        src = []
        mask_structure = []
        mask_sequence = []

        for seq, struct in zip(sequence, structure):
            choice = self.rng.choice(["seq", "struct", "x"], p=[0.4, 0.4, 0.2])

            if choice == "seq":
                trg_structure.append(struct)
                trg_sequence.append(self.blank_word)
                src.append(seq)
                mask_structure.append(1)
                mask_sequence.append(0)
            elif choice == "struct":
                trg_structure.append(self.blank_word)
                trg_sequence.append(seq)
                src.append(struct)
                mask_structure.append(0)
                mask_sequence.append(1)
            elif choice == "x":
                trg_structure.append(struct)
                trg_sequence.append(seq)
                src.append("X")
                mask_structure.append(1)
                mask_sequence.append(1)

        if self.rng.choice([True, False]):
            trg_joined = trg_structure + trg_sequence
            mask_joined = mask_structure + mask_sequence
        else:
            trg_joined = trg_sequence + trg_structure
            mask_joined = mask_sequence + mask_structure

        return src, trg_joined, mask_joined




if __name__ == "__main__":
    device = torch.device("cpu")
    batch_size = 12000

    data_dir = "/home/joerg/workspace/data/rna"

    # data_set = RNAct(data_dir=data_dir, max_len=200, min_len=10, source=[ "RNAStrAlign" ], include_N=True, design=False, partial=20)
    data_set = RNAct(data_dir=data_dir, max_len=1800, min_len=10, source=["RNA_Stand", "archiveII", "RNAStrAlign" , "bpRNA" ],
                     include_N=True, design=False, partial=False)
    # data_set = RNAct(data_dir=data_dir, max_len=2000, min_len=1, source=["RNA_Stand", "archiveII", "RNAStrAlign" ], include_N=False, design=False)



    train_iter = data_set.get_iterator("train", batch_size=batch_size, device=device, train=True)
    valid_iter = data_set.get_iterator("valid", batch_size=batch_size, device=device, train=False)
    test_iter = data_set.get_iterator("test", batch_size=batch_size, device=device, train=False)

    batch = next(train_iter.__iter__())
    src_data, src_length = batch.src
    trg_data, trg_length = batch.trg
    pos1id_data, pos1id_len = batch.pos1id

    # src_data  [batch size, sequencel length] [0,1,2,3,1,2,3,1] (int list)
    # stoi = data_set.SRC.vocab.stoi -> stoi == string to int is a dict {"<blank>":0, "A":1, "C":2, ... "N":}    stoi["A"]
    # itos = data_set.SRC.vocab.itos -> itos == int to string  is a list ["<blank>","A",...]    itos[0]
    # itos = data_set.TRG.vocab.itos -> itos == int to string  is a list ["<blank>",".", "(0",")0","(1",")1",...,")3"]    itos[0]

    print(f"# train set size: {data_set.set_size('train')}")
    seq_length = [batch.src[1].detach().cpu().numpy().tolist() for batch in train_iter]
    seq_length = [item for sublist in seq_length for item in sublist]
    # print("seq_length", [s.shape for s in seq_length if len(s.shape) > 1])

    print(f"# train seq length min: {np.min(seq_length)}")
    print(f"# train seq length mean: {np.mean(seq_length)}")
    print(f"# train seq length max: {np.max(seq_length)}")

    print(f"# valid set size: {data_set.set_size('valid')}")
    seq_length = [batch.src[1].detach().cpu().numpy().squeeze() for batch in valid_iter]
    seq_length = np.concatenate(seq_length, axis=0)
    print(f"# valid seq length min: {np.min(seq_length)}")
    print(f"# valid seq length mean: {np.mean(seq_length)}")
    print(f"# valid seq length max: {np.max(seq_length)}")

    print(f"# test set size: {data_set.set_size('test')}")

    print(f"src vocab: {data_set.src_vocab}")
    print(f"src vocab size: {data_set.src_vocab_size}")
    print(f"trg vocab: {data_set.trg_vocab}")
    print(f"trg vocab size: {data_set.trg_vocab_size}")



    # start_symbol = data_set.stoi(data_set.bos_word, origin='trg')
    # print("start_symbol: ", start_symbol)
    dist_list = []
    len_list = []
    knots_start_list = []
    for batch in train_iter:
        batch_pos1, len_pos1 = batch.pos1id
        batch_pos2, len_pos2 = batch.pos2id
        batch_src, len_src = batch.src

        assert torch.all(len_pos1.eq(len_pos2))

        for i in range(batch_pos1.size(0)):
            pos_id1 = batch_pos1[i,:len_pos1[i]].cpu().numpy()
            pos_id2 = batch_pos2[i,:len_pos2[i]].cpu().numpy()
            distance = np.abs(pos_id1-pos_id2)
            dist_list.append(distance)
            for dist, pos1, pos2 in zip(distance, pos_id1, pos_id2):
                knots_start_list.append([np.min([pos1, pos2]), dist])
            len_list.append(len_src[i])

    dist_list = np.concatenate(dist_list, 0)
    knots_start_list = np.asarray(knots_start_list)
    print("long_knots_start_list", knots_start_list.shape)
    import matplotlib.pyplot as plt
    plt.subplot(4, 1, 1)
    plt.hist(len_list, bins=40, label="sequence length")
    plt.ylabel("sequence length")
    plt.subplot(4, 1, 2)
    plt.hist(dist_list, bins=40, label="knot width")
    plt.ylabel("knot width")
    plt.subplot(4, 1, 3)
    plt.hist(knots_start_list[:,0], bins=40)
    plt.ylabel("knots_start")
    plt.subplot(4, 1, 4)
    plt.hist2d(knots_start_list[:, 1], knots_start_list[:, 0], bins=40)
    plt.ylabel("knots_start")
    plt.xlabel("knot width")
    plt.show()
    print("min pos dist: ", np.min(dist_list))
    print("mean pos dist: ", np.mean(dist_list))
    print("max pos dist: ", np.max(dist_list))

    # for batch in train_iter:
    #     target = batch.trg[0]
    #     target = target.cpu().numpy()
    #     target = target.flatten()
    #     uniques = np.unique(target).tolist()
    #     print(data_set.itos(uniques, origin="trg"))

    sample = train_iter.__iter__().__next__()
    #
    # print("src mat", sample.src[0].size())
    # print("src len", sample.src[1].size())
    #
    # print("trg mat", sample.trg[0].size())
    # print("trg len", sample.trg[1].size())
    #
    print("src sentence int", sample.src[0][0, :])
    print("src sentence int", sample.src[1][0])
    print("src sentence str", data_set.itos(sample.src[0][0, :].cpu().numpy().tolist(), origin="src"))

    print("trg sentence int", sample.trg[0][0, :])
    print("trg sentence int", sample.trg[1][0])
    print("trg sentence str", data_set.itos(sample.trg[0][0, :].cpu().numpy().tolist(), origin="trg"))
    #
    # print("pos1id", sample.pos1id[0][0, :])
    # print("pos2id", sample.pos2id[0][0, :])



from grakel import Graph
from grakel.kernels import (
    WeisfeilerLehman,
    VertexHistogram,
    #WeisfeilerLehmanOptimalAssignment,
    #ShortestPath,
    #Propagation,
    #RandomWalkLabeled,
    #CoreFramework,
)

from learna_tools.metrics.matrix_metrics import db2mat


def weisfeiler_lehmann(designed_sequence=None, designed_structure=None, target_sequence=None, target_structure=None):

    n_graph_iter = 5
    wl_kernel = WeisfeilerLehman(n_iter=n_graph_iter, normalize=True, base_graph_kernel=VertexHistogram)
    true_mat = db2mat(target_structure)
    pred_mat = db2mat(designed_structure)
    true_graph = Graph(initialization_object=true_mat.astype(int), node_labels={s:str(s) for s in range(true_mat.shape[0])})
    pred_graph = Graph(initialization_object=pred_mat.astype(int), node_labels={s:str(s) for s in range(pred_mat.shape[0])})

    wl_kernel.fit_transform([true_graph])
    wl_pred = wl_kernel.transform([pred_graph])

    return 1 - wl_pred[0][0]



import time
from pathlib import Path
from itertools import islice
from RNA import fold


class FamilyObject(object):
    def __init__(self, family_path):
        print(f"Create family object of rfam family {family_path.stem}")
        self._path = family_path
        self._id = family_path.stem
        self._content = family_path.read_text().splitlines()

    def __iter__(self):
        targets = self.targets
        return iter(targets)

    def _get_targets(self):
        return [(self.id, len(sequence), sequence.upper().replace('T', 'U'), fold(sequence), ((sequence.upper().count('G') + sequence.upper().count('C')) / len(sequence))) for sequence in islice(self._content, 1, len(self._content), 2)]

    @property
    def id(self):
        return self._id

    @property
    def targets(self):
        return self._get_targets()


def read_family_targets(database_dir, family):
    if family:
        try:
            return [FamilyObject(Path(database_dir, f"RF{family}.fa"))]
        except:
            print(f"family {family} doesn't exist")
    families = [FamilyObject(family_path) for family_path in Path(database_dir).glob('*.fa')]
    return iter(families)

def database_to_interim(database_dir, family, out_dir):
    out_path = Path(out_dir, 'rfam_database.tsv')
    write_database(database_dir, family, out_path)

def write_database(database_dir, family, out_path):
    print("Start writing database file")
    if family:
        out_path = Path(out_path.parent, f"targets_{family}.tsv")
    database_file = open(out_path, "a+")
    database_file.write("family\tlength\tsequence\tstructure_and_mfe\tgc_content\n")
    for family in read_family_targets(database_dir, family):
        print(f"Writing targets of family {family.id} into database")
        for target in family:
            database_file.write(f"{target[0]}\t{target[1]}\t{target[2]}\t{target[3]}\t{target[4]}\n")
    database_file.close()
    print(f"Database written to {out_path.resolve()}")


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()

    parser.add_argument("--data", default="data/rfam_raw/", type=Path, help="Output path")
    parser.add_argument("--out_dir", default="data/test/", help="Output path")
    parser.add_argument(
        "--family",
        default=None,
        required=False,
        type=str,
        help="Rfam family id to use for dataset. Requires all ",
    )

    args = parser.parse_args()

    out_dir = args.out_dir
    data = args.data
    family = ("0" * (5 - len(args.family))) + args.family


    start = time.time()
    database_to_interim(data, family, out_dir)
    end = time.time()
    print(f"dataset generated in {end - start} seconds")

    # print(pd.read_csv(Path(out_dir, 'rfam_database.tsv'), sep='\t'))



# Copyright 2017 reinforce.io. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

# Changes from original tensorforce version include: restart capability and making
# the weight updates optional.


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import time
import tensorflow as tf
from six.moves import xrange


class Runner(object):
    """
    Simple runner for non-realtime single-process execution.
    """

    def __init__(self, get_agent, environment, repeat_actions=1, history=None):
        """
        Initialize a Runner object.

        Args:
            agent:
            environment:
            repeat_actions:
        """
        self.get_agent = get_agent
        self.agent = get_agent()
        self.environment = environment
        self.repeat_actions = repeat_actions

        self.reset(history)

    def reset(self, history=None):
        # If history is empty, use default values in history.get().
        if not history:
            history = dict()

        self.episode_rewards = history.get("episode_rewards", list())
        self.episode_timesteps = history.get("episode_timesteps", list())
        self.episode_times = history.get("episode_times", list())

    def run(
        self,
        timesteps=None,
        episodes=None,
        restart_timeout=None,
        max_episode_timesteps=None,
        stop_learning=False,
        deterministic=False,
        episode_finished=None,
    ):
        """
        Runs the agent on the environment.

        Args:
            timesteps: Number of timesteps
            episodes: Number of episodes
            max_episode_timesteps: Max number of timesteps per episode
            deterministic: Deterministic flag
            episode_finished: Function handler taking a `Runner` argument and returning a boolean indicating
                whether to continue execution. For instance, useful for reporting intermediate performance or
                integrating termination conditions.
        """

        # Keep track of episode reward and episode length for statistics.
        self.start_time = time.time()

        self.agent.reset()

        self.episode = self.agent.episode
        if episodes is not None:
            episodes += self.agent.episode

        self.timestep = self.agent.timestep
        if timesteps is not None:
            timesteps += self.agent.timestep

        iteration_start = time.time()
        while True:
            episode_start_time = time.time()

            self.agent.reset()
            state = self.environment.reset()
            episode_reward = 0
            self.episode_timestep = 0

            while True:
                action = self.agent.act(states=state, deterministic=deterministic)

                if self.repeat_actions > 1:
                    reward = 0
                    for repeat in xrange(self.repeat_actions):
                        state, terminal, step_reward = self.environment.execute(
                            actions=action
                        )
                        reward += step_reward
                        if terminal:
                            break
                else:
                    state, terminal, reward = self.environment.execute(actions=action)

                if (
                    max_episode_timesteps is not None
                    and self.episode_timestep >= max_episode_timesteps
                ):
                    terminal = True

                if not stop_learning:
                    self.agent.observe(terminal=terminal, reward=reward)

                self.episode_timestep += 1
                self.timestep += 1
                episode_reward += reward

                if (
                    terminal or self.agent.should_stop()
                ):  # TODO: should_stop also termina?
                    break

            time_passed = time.time() - episode_start_time

            self.episode_rewards.append(episode_reward)
            self.episode_timesteps.append(self.episode_timestep)
            self.episode_times.append(time_passed)

            self.episode += 1

            if (
                episode_finished
                and not episode_finished(self)
                or (episodes is not None and self.agent.episode >= episodes)
                or (timesteps is not None and self.agent.timestep >= timesteps)
                or self.agent.should_stop()
            ):
                # agent.episode / agent.timestep are globally updated
                break

            iteration_time = time.time() - iteration_start
            if restart_timeout and restart_timeout < iteration_time:
                print("restarting")
                self.agent.close()
                self.agent = self.get_agent()
                iteration_start = time.time()

        self.agent.close()
        self.environment.close()



from pathlib import Path
import numpy as np

def _encode_pairing(secondary):
    """
    Encode the base pairing scheme of the design task.

    Args:
        secondary: the secondary structure (domain) that has to get encoded.

    Returns:
        List that contains the paired site for each site if exists, or None otherwise.
    """
    pairing_encoding = [None] * len(secondary)
    stack = []
    for index, symbol in enumerate(secondary, 0):
        if symbol == "(":
            stack.append(index)
        elif symbol == ")":
            if not stack:
                continue
            else:
                paired_site = stack.pop()
                pairing_encoding[paired_site] = index
                pairing_encoding[index] = paired_site
    return pairing_encoding

def random_insert_n(string):
    string_list = list(string)

    if string_list[0] in ['A', 'C', 'G', 'U']:
        insert_times = np.random.randint(1, 8)
    else:
        insert_times = np.random.randint(0, 3)

    percentage = 0.3
    length_change = np.random.random_sample((insert_times, )) * percentage
    inserts = ['N' * (int(len(string) * l)) for l in length_change]
    if insert_times != 0:
        max_insert_position = len(string) - np.max([(int(len(string) * l)) for l in length_change])
        insert_positions = np.random.choice(max_insert_position, insert_times)
        for insert, position in zip(inserts, insert_positions):
            string_list[position:position + len(insert)] = insert

    return ''.join(string_list)


def remove_unbalanced_brackets(secondary, mode):
    secondary_list = list(secondary.upper())
    stack = []
    for index, symbol in enumerate(secondary, 0):
        if symbol == "(":
            stack.append(index)
        elif symbol == ")":
            if not stack:
                secondary_list[index] = 'N'
            else:
                stack.pop()
        elif symbol in ['A', 'C', 'G', 'U']:
            if mode == 'anta':
                secondary_list[index] = 'N'
    for index in stack:
        secondary_list[index] = 'N'
    return ''.join(secondary_list)

def extend_unbalanced2balanced(secondary, pairing_encoding, mode):
    # unbalanced_closing = []
    # stack = []
    # for index, symbol in enumerate(secondary, 0):
    #     if symbol == "(":
    #         stack.append(index)
    #     elif symbol == ")":
    #         if not stack:
    #             unbalanced_closing.append(index)
    #             # secondary_list[index] = 'N'
    #         else:
    #             stack.pop()
    #     elif symbol in ['A', 'C', 'G', 'U']:
    #         secondary_list[index] = 'N'
    # for index in stack:
    #     secondary_list[pairing_encoding[index]] = ')'
    # for index in unbalanced_closing:
    #     secondary_list[pairing_encoding[index]] = '('
    secondary_list = list(secondary.upper())
    for index, symbol in enumerate(secondary_list, 0):
        if symbol == "(":
            secondary_list[pairing_encoding[index]] = ')'
        elif symbol == ")":
            secondary_list[pairing_encoding[index]] = '('
        elif symbol in ['A', 'C', 'G', 'U']:
            if mode == 'anta':
                secondary_list[index] = 'N'
    return ''.join(secondary_list)

def verify_balance(secondary):
    secondary_list = list(secondary.upper())
    stack = []
    for index, symbol in enumerate(secondary, 0):
        if symbol == "(":
            stack.append(index)
        elif symbol == ")":
            if not stack:
                return False
            else:
                stack.pop()
        # elif symbol in ['A', 'C', 'G', 'U']:
        #     secondary_list[index] = 'N'
    # for index in stack:
    #     secondary_list[index] = 'N'
    return not bool(stack)

def get_unbalanced_brackets(secondary):
    secondary_list = list(secondary.upper())
    unbalanced_closing = []
    stack = []
    for index, symbol in enumerate(secondary, 0):
        if symbol == "(":
            stack.append(index)
        elif symbol == ")":
            if not stack:
                unbalanced_closing.append(index)
            else:
                stack.pop()
        # elif symbol in ['A', 'C', 'G', 'U']:
        #     secondary_list[index] = 'N'
    # for index in stack:
    #     secondary_list[index] = 'N'
    for index in stack:
        secondary_list[index] = 'L'
    for index in unbalanced_closing:
        secondary_list[index] = 'R'
    return ''.join(secondary_list)

def make_local_data(secondary, primary):
    primary_list = list(primary)
    for index, symbol in enumerate(secondary, 0):
        if symbol in ['.', '(', ')']:
            primary_list[index] = 'N'
    return ''.join(primary_list)



if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data_path",
        default="data/rfam_local_test",
        help="Path to data directory",
    )
    parser.add_argument(
        "--dataset",
        default="rfam_local_test",
        help="Path to data directory",
    )

    args = parser.parse_args()

    dataset_path = Path(args.data_path, args.dataset)
    # out_dir = 'data/rfam_local_test_anta/'
    # out_dir_path = Path(out_dir)
    # out_dir_path.mkdir(exist_ok=True, parents=True)
    # counter = 0
    # length = []
    strucs = []
    counter = 0
    for path in dataset_path.glob('*.rna'):
        # id = path.read_text().split('\t')[0]
        # struc = path.read_text().split('\t')[1]
        # seq = path.read_text().split('\t')[2]
        # local = path.read_text().split('\t')[3]
        # gc = path.read_text().split('\t')[5]
        # mfe = path.read_text().split('\t')[6]
        # print(path.read_text().split())
        seq = path.read_text().split()[2].strip()
        struc = path.read_text().split()[1].strip()
        gc = path.read_text().split()[3].strip()
        id = path.stem
        # gc = path.stem
        # mfe = path.stem
        # seq = 'N' * len(struc)

        # print(id)
        # struc = local.replace('A', 'N').replace('C', 'N').replace('G', 'N').replace('U', 'N')
        # seq = local.replace('.', 'N').replace('(', 'N').replace(')', 'N')
        print(struc)
        print(seq)
        print(gc)
        # print(gc)
        # print(mfe)
        # print()

        out_dir = Path('archiveII_thirdparty')
        out_dir.mkdir(exist_ok=True, parents=True)

        out = Path(out_dir, f"{id}.rna")

        with open(out, 'w+') as f:
            f.write(struc + '\t' + seq + '\t' + gc)


        # motif_based = path.read_text().split('\t')[3]
        # gc = path.read_text().split('\t')[4]
        # assert len(seq) == len(struc)

        # length.append(len(seq))
        # strucs.append(struc)

        # if int(path.stem)%3 == 0:
        #     struc_c = random_insert_n(struc)
        #     seq_c = random_insert_n(seq)
        #     while seq_c.count('N') / len(seq_c) < 0.4:
        #         seq_c = random_insert_n(seq_c)
        #     # print(path.stem)
        #     # print(struc_c)
        #     # print(seq_c)
        # else:
        #     struc_c = local.replace('A', 'N').replace('C', 'N').replace('G', 'N').replace('U', 'N')
        #     seq_c = local.replace('.', 'N').replace('(', 'N').replace(')', 'N')
        #     # print(path.stem)
        #     # print(struc_c)
        #     # print(seq_c)
        # if not 'N' in seq_c:
        #     seq_c = random_insert_n(seq_c)
        # print(path.stem)
        # print(struc_c)
        # print(seq_c)
        # out_path = Path(args.data_path + '/' + args.dataset + '_new/', f"{path.stem}.rna")
        # out_path.parent.mkdir(exist_ok=True)
        # out_path.write_text(f"{path.stem}\t{struc_c}\t{seq_c}\t{gc}\t{mfe}")



        # seq_i = random_insert_n(seq)
        # struc_i = random_insert_n(struc)
        # if struc_i[-1] == 'N' or seq_i[-1] == 'N':
        #     print(seq_i)
        #     print(struc_i)




        # counter += 1
        # if counter == 20:
        #     break
        # counter += 1
    # avg_length = length / counter
    # length = np.asarray(length)
    # strucs = np.asarray(strucs)
    # length = np.array(list(map(len, strucs)))
    # avg_length = np.mean(length)

    # print(f"average length: {avg_length}")
    # print(f"max length: {np.max(length)}")
    # print(f"min length: {np.min(length)}")
    # print(strucs[np.where(length == 104)])
    # print('N' * 104)




        # print(struc)
        # balanced_struc_remove = remove_unbalanced_brackets(local, 'anta')
        # balanced_struc_remove_learna = remove_unbalanced_brackets(local, 'learna')
        # pairing_encoding = _encode_pairing(struc)
        # balanced_struc_add = extend_unbalanced2balanced(local, pairing_encoding, 'anta')
        # balanced_struc_add_learna = extend_unbalanced2balanced(local, pairing_encoding, 'learna')

        # print(local)
        # print(balanced_struc_remove)
        # print(balanced_struc_add)
        # print(balanced_struc_remove_learna)
        # print(balanced_struc_add_learna)
        # print(struc)
        # seq_c = local.replace('.', 'N').replace('(', 'N').replace(')', 'N')
        # seq_c = make_local_data(balanced_struc_add, seq_c)
        # print('\n')
        # print(balanced_struc_remove_learna)
        # print(balanced_struc_remove)
        # print(balanced_struc_add)
        # print(struc)
        # print(seq)
        # print(seq_c)
        # print(local)
        # print(gc)
        # if not verify_balance(balanced_struc_add) or not verify_balance(balanced_struc_remove):
        #     print(get_unbalanced_brackets(balanced_struc_add))
        #     print(get_unbalanced_brackets(balanced_struc_remove))
        # print(seq_c)
        # print(verify_balance(balanced_struc_add), verify_balance(balanced_struc_remove))
        # out_path_anta = Path(args.data_path + '/' + args.dataset + '_balanced/', f"{path.stem}.anta")
        # out_path_learna = Path(args.data_path + '/' + args.dataset + '_balanced/', f"{path.stem}.rna")
        # out_path_anta.parent.mkdir(exist_ok=True)
        # out_path_learna.parent.mkdir(exist_ok=True)
        # print(out_path_anta.resolve())

        # out_path_anta_add = Path( + '_balanced', f"{path.stem}.anta")
        #out_path_anta.write_text(f"{balanced_struc_add}\t{seq_c}")
        #out_path_learna.write_text(f"{path.stem}\t{balanced_struc_add}\t{seq_c}\t{path.stem}\t{path.stem}")

        # out_path_anta_remove = Path(dataset_path + '_balanced_remove', f"{path.stem}.anta")
        # out_path_anta_remove.write_text(f"{balanced_struc_remove}\t{seq_c}")


        # out_path_learna_add = Path(dataset_path + '_balanced_add', f"{path.stem}.anta")
        # out_path_learna_add.write_text(f"{balanced_struc_add}\t{seq}")

        # out_path_learna_remove = Path(dataset_path + '_balanced_remove', f"{path.stem}.anta")
        # out_path_learna_remove.write_text(f"{path.stem}\t{struc}\t{seq}{balanced_struc_remove}\t{seq}")



import subprocess
import datetime
import numpy as np
import pandas as pd

from pathlib import Path
from collections import defaultdict

import matplotlib.pyplot as plt
import matplotlib.image as mpimg

# from Bio.Seq import Seq
# from Bio import motifs
import logomaker as lm


def display_varna_plot(path):
    image = mpimg.imread(path)
    plt.imshow(image)
    plt.axis("off")   # turns off axes
    plt.axis("tight")  # gets rid of white border
    plt.axis("image")  # square up the image instead of filling the "figure" space
    plt.show()



def plot_with_varna(pairs, sequence, Id, plot_dir='plots', plot_type='radiate', resolution='2.0'):
        Path(plot_dir).mkdir(exist_ok=True, parents=True)
        multiplets = get_multiplets(pairs)
        watson_pairs, wobble_pairs, noncanonical_pairs = type_pairs(pairs, sequence)
        lone_bp = lone_pair(pairs)
        tertiary_bp = multiplets + noncanonical_pairs + lone_bp
        tertiary_bp = [list(x) for x in set(tuple(x) for x in tertiary_bp)]

        str_tertiary = []
        for i,I in enumerate(tertiary_bp):
            if i==0:
                str_tertiary += ('(' + str(I[0]+1) + ',' + str(I[1]+1) + '):color=""#FFFF00""')
            else:
                str_tertiary += (';(' + str(I[0]+1) + ',' + str(I[1]+1) + '):color=""#FFFF00""')

        tertiary_bp = ''.join(str_tertiary)

        ct_path = Path(plot_dir, f"{Id}.ct")

        ct_file_output(pairs, sequence, Id, ct_path)

        varna_path = str(Path(plot_dir, f"{Id}").resolve())
        if plot_type == 'radiate':
            subprocess.Popen(["varna", '-i', str(ct_path.resolve()), '-o', varna_path + '_radiate.png', '-algorithm', 'radiate', '-resolution', resolution, '-bpStyle', 'lw', '-auxBPs', tertiary_bp], stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]  # -> to plot structure only: '-basesStyle1', 'fill=#FFFFFF,label=#FFFFFF,number=#FFFFFF', '-applyBasesStyle1on', ','.join([str(i) for i in range(1, len(sequence)+1)]),
        elif plot_type == 'line':
            subprocess.Popen(["varna", '-i', str(ct_path.resolve()), '-o', varna_path + '_line.png', '-algorithm', 'line', '-resolution', resolution, '-bpStyle', 'lw', '-auxBPs', tertiary_bp], stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]



def plot_df_with_varna(df, plot_dir='plots', name='', plot_type='radiate', resolution='2.0', show=False):
    if name:
        plot_dir = Path(plot_dir, name)
    else:
        current_time = f'{datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")}'
        plot_dir = Path(plot_dir, current_time)
    Ids = defaultdict(list)
    for i, row in df.iterrows():
        pairs = db2pairs(row['structure'])
        sequence = row['sequence']
        Ids[str(row['Id'])].append(str(row['Id']))
        Id = str(row['Id']) + '_' + str(len(Ids[str(row['Id'])])-1)
        plot_with_varna(pairs, sequence, Id, plot_dir=plot_dir, plot_type=plot_type, resolution=resolution)
        if show:
            display_varna_plot(Path(plot_dir, f"{Id}" + '_' + f"{plot_type}.png").resolve())


def plot_weblogo(df):
    """
    Seems like the weblogo option of Biopython is not working properly...
    This function simply does not work.
    """
    df['sequence'] = df['sequence'].apply(lambda x: x.replace('U', 'T'))
    instances = df['sequence']
    m = motifs.create(instances)
    m.weblogo('logo.png')

def plot_sequence_logo(df, plotting_dir='plots', name='current_sequence_logo', show=False):
    """
    Create a logo from a dataframe that contains a 'sequence' column.

    All sequences have to be of the same length.
    """

    outdir = Path(plotting_dir, 'logos')
    outdir.mkdir(exist_ok=True, parents=True)
    sequences = df['sequence'].to_list()
    counts_mat = lm.alignment_to_matrix(sequences) / len(sequences)
    counts_mat.head()
    logo = lm.Logo(counts_mat, stack_order='small_on_top', figsize=[30, 6])  # , vsep=0.05, stack_order='small_on_top')
    logo.style_spines(visible=False)
    logo.style_spines(spines=['left','bottom', 'top', 'right'], visible=True, linewidth=2)
    
    # Hide the x-axis labels and ticks
    logo.ax.set_xticks([])
    logo.ax.set_xticklabels([])
    
    # Hide the y-axis labels and ticks
    logo.ax.set_yticks([])
    logo.ax.set_yticklabels([])
    
    save_path = Path(outdir, f"{name}.png")
    plt.savefig(save_path)
    if show:
        plt.show()
    plt.close('all')


def plot_structure_logo(df, plotting_dir='plots', name='current_structure_logo', show=False):
    """
    Create a logo from a dataframe that contains a 'structure' column.

    All structures have to be of the same length.
    """
    outdir = Path(plotting_dir, 'logos')
    outdir.mkdir(exist_ok=True, parents=True)
    sequences = df['structure'].to_list()
    counts_mat = lm.alignment_to_matrix(sequences) / len(sequences)
    counts_mat.head()
    logo = lm.Logo(counts_mat, stack_order='small_on_top')  # , vsep=0.05, stack_order='small_on_top')
    logo.style_spines(visible=False)
    logo.style_spines(spines=['left','bottom'], visible=True, linewidth=2)
    save_path = Path(outdir, f"{name}.png")
    plt.savefig(save_path)
    if show:
        plt.show()
    plt.close('all')


################################################################################
# following code snippets originate from SPOT-RNA repo at https://github.com/jaswindersingh2/SPOT-RNA
# we use them for plotting with varna in the visualization module
################################################################################


# copy-paste from SPOT-RNA2 source code
def lone_pair(pairs):
    lone_pairs = []
    pairs.sort()
    for i, I in enumerate(pairs):
        if ([I[0] - 1, I[1] + 1] not in pairs) and ([I[0] + 1, I[1] - 1] not in pairs):
            lone_pairs.append(I)

    return lone_pairs


# copy-paste from SPOT-RNA2 source code
def type_pairs(pairs, sequence):
    sequence = [i.upper() for i in sequence]
    # seq_pairs = [[sequence[i[0]],sequence[i[1]]] for i in pairs]

    AU_pair = []
    GC_pair = []
    GU_pair = []
    other_pairs = []
    for i in pairs:
        if [sequence[i[0]],sequence[i[1]]] in [["A","U"], ["U","A"]]:
            AU_pair.append(i)
        elif [sequence[i[0]],sequence[i[1]]] in [["G","C"], ["C","G"]]:
            GC_pair.append(i)
        elif [sequence[i[0]],sequence[i[1]]] in [["G","U"], ["U","G"]]:
            GU_pair.append(i)
        else:
            other_pairs.append(i)
    watson_pairs_t = AU_pair + GC_pair
    wobble_pairs_t = GU_pair
    other_pairs_t = other_pairs
        # print(watson_pairs_t, wobble_pairs_t, other_pairs_t)
    return watson_pairs_t, wobble_pairs_t, other_pairs_t



# copy-paste from SPOT-RNA2 source code
def multiplets_pairs(pred_pairs):

    pred_pair = [i[:2] for i in pred_pairs]
    temp_list = flatten(pred_pair)
    temp_list.sort()
    new_list = sorted(set(temp_list))
    dup_list = []
    for i in range(len(new_list)):
        if (temp_list.count(new_list[i]) > 1):
            dup_list.append(new_list[i])

    dub_pairs = []
    for e in pred_pair:
        if e[0] in dup_list:
            dub_pairs.append(e)
        elif e[1] in dup_list:
            dub_pairs.append(e)

    temp3 = []
    for i in dup_list:
        temp4 = []
        for k in dub_pairs:
            if i in k:
                temp4.append(k)
        temp3.append(temp4)

    return temp3


# copy-paste from SPOT-RNA2 source code
def multiplets_free_bp(pred_pairs, y_pred):
    L = len(pred_pairs)
    multiplets_bp = multiplets_pairs(pred_pairs)
    save_multiplets = []
    while len(multiplets_bp) > 0:
        remove_pairs = []
        for i in multiplets_bp:
            save_prob = []
            for j in i:
                save_prob.append(y_pred[j[0], j[1]])
            remove_pairs.append(i[save_prob.index(min(save_prob))])
            save_multiplets.append(i[save_prob.index(min(save_prob))])
        pred_pairs = [k for k in pred_pairs if k not in remove_pairs]
        multiplets_bp = multiplets_pairs(pred_pairs)
    save_multiplets = [list(x) for x in set(tuple(x) for x in save_multiplets)]
    assert L == len(pred_pairs)+len(save_multiplets)
    #print(L, len(pred_pairs), save_multiplets)
    return pred_pairs, save_multiplets


def flatten(x):
    result = []
    for el in x:
        if hasattr(el, "__iter__") and not isinstance(el, str):
            result.extend(flatten(el))
        else:
            result.append(el)
    return result

def ct_file_output(pairs, seq, id, save_result_path):

    col1 = np.arange(1, len(seq) + 1, 1)
    col2 = np.array([i for i in seq])
    col3 = np.arange(0, len(seq), 1)
    col4 = np.append(np.delete(col1, 0), [0])
    col5 = np.zeros(len(seq), dtype=int)

    for i, I in enumerate(pairs):
        col5[I[0]] = int(I[1]) + 1
        col5[I[1]] = int(I[0]) + 1
    col6 = np.arange(1, len(seq) + 1, 1)
    temp = np.vstack((np.char.mod('%d', col1), col2, np.char.mod('%d', col3), np.char.mod('%d', col4),
                      np.char.mod('%d', col5), np.char.mod('%d', col6))).T
    #os.chdir(save_result_path)
    #print(os.path.join(save_result_path, str(id[0:-1]))+'.spotrna')
    np.savetxt(save_result_path, (temp), delimiter='\t\t', fmt="%s", header=str(len(seq)) + '\t\t' + str(id) + '\t\t' + ' output\n' , comments='')  # changed outpath from original SPOT-RNA code

    return


def get_multiplets(pairs):
    pair_dict = defaultdict(list)
    for p in pairs:
        pair_dict[p[0]].append(p[1])
        pair_dict[p[1]].append(p[0])

    multiplets = []

    for k, v in pair_dict.items():
        if len(v) > 1:
            for p in v:
                multiplets.append(tuple(sorted([k, p])))
    multiplets = list(set(multiplets))

    return multiplets

################################################################################

def db2pairs(structure, start_index=0):
    """
    Converts dot-bracket string into a list of pairs.

    Input:
      structure <string>: A sequence in dot-bracket format.
      start_index <int>: Starting index of first nucleotide (default zero-indexing).

    Returns:
      pairs <list>: A list of tuples of (index1, index2, pk_level).

    """
    level_stacks = defaultdict(list)
    closing_partners = {')': '(', ']': '[', '}': '{', '>': '<'}
    levels = {')': 0, ']': 1, '}': 2, '>': 3}

    pairs = []

    for i, sym in enumerate(structure, start_index):
        if sym == '.':
            continue
        # high order pks are alphabetical characters
        if sym.isalpha():
            if sym.isupper():
                level_stacks[sym].append(i)
            else:
                try:  # in case we have invalid preditions, we continue with next bracket
                    op = level_stacks[sym.upper()].pop()
                    pairs.append((op, i,
                                  ord(sym.upper()) - 61))  # use asci code if letter is used to asign PKs, start with level 4 (A has asci code 65)
                except:
                    continue
        else:
            if sym in closing_partners.values():
                level_stacks[sym].append(i)
            else:
                try:  # in case we have invalid preditions, we continue with next bracket
                    op = level_stacks[closing_partners[sym]].pop()
                    pairs.append([op, i, levels[sym]])
                except:
                    continue
    return sorted(pairs, key=lambda x: x[0])



import os

os.environ["OMP_NUM_THREADS"] = "1"
# os.environ['KMP_AFFINITY']='compact,1,0'

import multiprocessing

import numpy as np
import ConfigSpace as CS
from pathlib import Path
from hpbandster.core.worker import Worker


from src.learna.agent import NetworkConfig, get_network, AgentConfig
from src.learna.environment import RnaDesignEnvironment, RnaDesignEnvironmentConfig
from src.learna.design_rna import design_rna
from src.data.parse_dot_brackets import parse_dot_brackets, parse_local_design_data


class LearnaWorker(Worker):
    def __init__(self, data_dir, num_cores, train_sequences, **kwargs):
        super().__init__(**kwargs)
        self.num_cores = num_cores
        self._data_dir = data_dir
        self.sequence_ids = train_sequences
        # self.train_sequences = parse_dot_brackets(
        #     dataset="rfam_local_validation",
        #     data_dir=data_dir,
        #     target_structure_ids=train_sequences,
        # )

    def compute(self, config, budget, **kwargs):
        """
		Parameters
		----------
			budget: float
				cutoff for the agent on a single sequence
		"""

        config = self._fill_config(config)

        if config["local_design"]:
            self.train_sequences = parse_local_design_data(
            dataset=Path(self._data_dir, "rfam_local_validation").stem,
            data_dir=self._data_dir,
            target_structure_ids=self.sequence_ids,
            )
        else:
            self.train_sequences = parse_dot_brackets(
                dataset=Path(self._data_dir, "rfam_local_validation").stem,
                data_dir=self._data_dir,
                target_structure_ids=self.sequence_ids,
            )

        network_config = NetworkConfig(
            conv_sizes=[config["conv_size1"], config["conv_size2"]],
            conv_channels=[config["conv_channels1"], config["conv_channels2"]],
            num_fc_layers=config["num_fc_layers"],
            fc_units=config["fc_units"],
            num_lstm_layers=config["num_lstm_layers"],
            lstm_units=config["lstm_units"],
            embedding_size=config["embedding_size"],
        )

        agent_config = AgentConfig(
            learning_rate=config["learning_rate"],
            batch_size=config["batch_size"],
            entropy_regularization=config["entropy_regularization"],
        )

        env_config = RnaDesignEnvironmentConfig(
            reward_exponent=config["reward_exponent"],
            state_radius=config["state_radius"],
            local_design=config["local_design"],
            # sequence_reward=bool(config["sequence_reward"]),
            reward_function=config["reward_function"],
            predict_pairs=config["predict_pairs"],
            state_representation=config["state_representation"],
            data_type=config["data_type"],
            # gc_weight=config["gc_weight"],
            # structural_weight=config["structural_weight"],
            # desired_gc=config["desired_gc"],
            # gc_improvement_step=True,
            # gc_postprocessing=False,
        )

        validation_info = self._evaluate(
            budget, config["restart_timeout"], network_config, agent_config, env_config
        )

        return {
            "loss": validation_info["num_unsolved"],
            # "loss": validation_info["sum_of_min_distances"],
            # "loss": validation_info["sum_of_min_deltas_and_distances"],
            "info": {"validation_info": validation_info},
        }

    def _evaluate(
        self,
        evaluation_timeout,
        restart_timeout,
        network_config,
        agent_config,
        env_config,
    ):

        evaluation_arguments = [
            [
                [train_sequence],
                evaluation_timeout,  # timeout
                None,  # restore_path
                False,  # stop_learning
                restart_timeout,  # restart_timeout
                network_config,
                agent_config,
                env_config,
            ]
            for train_sequence in self.train_sequences
        ]

        with multiprocessing.Pool(self.num_cores) as pool:
            evaluation_results = pool.starmap(design_rna, evaluation_arguments)

        evaluation_sequence_infos = {}
        evaluation_sum_of_min_distances = 0
        evaluation_sum_of_first_distances = 0
        evaluation_num_solved = 0
        evaluation_num_unsolved = 0
        # evaluation_sum_of_min_gc_deltas = 0
        # evaluation_sum_of_min_gc_deltas_and_distances = 0

        for r in evaluation_results:
            sequence_id = r[0].target_id
            r.sort(key=lambda e: e.time)

            times = np.array(list(map(lambda e: e.time, r)))
            dists = np.array(list(map(lambda e: e.normalized_hamming_distance, r)))
            # deltas_gc = np.array(list(map(lambda e: e.delta_gc, r)))
            # deltas_and_distances = np.array(list(map(lambda e: e.delta_gc + e.normalized_hamming_distance, r)))

            evaluation_sum_of_min_distances += dists.min()
            evaluation_sum_of_first_distances += dists[0]

            # evaluation_sum_of_min_gc_deltas += deltas_gc.min()
            # evaluation_sum_of_min_gc_deltas_and_distances += deltas_and_distances.min()

            evaluation_num_solved += dists.min() == 0.0
            evaluation_num_unsolved += dists.min() != 0

            evaluation_sequence_infos[sequence_id] = {
                "num_episodes": len(r),
                "mean_time_per_episode": float((times[1:] - times[:-1]).mean()),
                "min_distance": float(dists.min()),
                "last_distance": float(dists[-1]),
                # "min_delta_gc": float(deltas_gc.min()),
                # "last_delta_gc": float(deltas_gc[-1]),
                # "min_deltas_and_distances": float(deltas_and_distances.min()),
                # "last_deltas_and_distances": float(deltas_and_distances[-1]),
            }

        evaluation_info = {
            "num_solved": int(evaluation_num_solved),
            "num_unsolved": int(evaluation_num_unsolved),
            "sum_of_min_distances": float(evaluation_sum_of_min_distances),
            "sum_of_first_distances": float(evaluation_sum_of_first_distances),
            # "sum_of_min_gc_deltas": float(evaluation_sum_of_min_gc_deltas),
            # "sum_of_min_deltas_and_distances": float(evaluation_sum_of_min_gc_deltas_and_distances),
            "squence_infos": evaluation_sequence_infos,
        }

        return evaluation_info

    @staticmethod
    def get_configspace():
        config_space = CS.ConfigurationSpace()

        # parameters for PPO here
        config_space.add_hyperparameter(
            CS.UniformFloatHyperparameter(
                "learning_rate", lower=1e-6, upper=1e-3, log=True, default_value=5e-4  # FR: changed learning rate lower from 1e-5 to 1e-6, ICLR: Learna (5,99e-4), Meta-LEARNA (6.44e-5)
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "batch_size", lower=32, upper=256, log=True, default_value=32  # FR: changed batch size upper from 128 to 256, configs from ICLR used 126 (LEARNA) and 123 (Meta-LEARNA)
            )
        )
        config_space.add_hyperparameter(
            CS.UniformFloatHyperparameter(
                "entropy_regularization",
                lower=1e-7,  # FR: changed entropy regularization lower from 1e-5 to 1e-7, ICLR: LEARNA (6,76e-5), Meta-LEARNA (151e-4)
                upper=1e-2,
                log=True,
                default_value=1.5e-3,
            )
        )

        config_space.add_hyperparameter(
            CS.UniformFloatHyperparameter(
                "reward_exponent", lower=1, upper=12, default_value=1  # FR: changed reward_exponent upper from 10 to 12, ICLR: Learna (9.34), Meta-LEARNA (8.93)
            )
        )

        config_space.add_hyperparameter(
            CS.UniformFloatHyperparameter(
                "state_radius_relative", lower=0, upper=1, default_value=0
            )
        )

        # parameters for the architecture
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "conv_radius1", lower=0, upper=8, default_value=1
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "conv_channels1", lower=1, upper=32, log=True, default_value=32
            )
        )

        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "conv_radius2", lower=0, upper=4, default_value=0
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "conv_channels2", lower=1, upper=32, log=True, default_value=1
            )
        )

        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "num_fc_layers", lower=1, upper=2, default_value=2
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "fc_units", lower=8, upper=64, log=True, default_value=50
            )
        )

        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "num_lstm_layers", lower=0, upper=3, default_value=0  # FR: changed lstm layers upper from 2 to 3
            )
        )
        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "lstm_units", lower=1, upper=64, log=True, default_value=1
            )
        )

        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "embedding_size", lower=0, upper=8, default_value=1  # FR: changed embedding size upper from 4 to 8
            )
        )

        # config_space.add_hyperparameter(
        #     CS.UniformIntegerHyperparameter(
        #         "sequence_reward", lower=0, upper=1, default_value=0
        #     )
        # )

        config_space.add_hyperparameter(
            CS.CategoricalHyperparameter(
                "reward_function", choices=['sequence_and_structure', 'structure_replace_sequence', 'structure_only']
            )
        )

        config_space.add_hyperparameter(
            CS.CategoricalHyperparameter(
                "state_representation", choices=['n-gram', 'sequence_progress']
            )
        )

        config_space.add_hyperparameter(
            CS.CategoricalHyperparameter(
                "data_type", choices=['random', 'random-sort']
            )
        )


        config_space.add_hyperparameter(
            CS.UniformIntegerHyperparameter(
                "predict_pairs", lower=0, upper=1, default_value=1
            )
        )


        # config_space.add_hyperparameter(
        #     CS.UniformFloatHyperparameter(
        #         "structural_weight", lower=0, upper=1, default_value=1
        #     )
        # )

        # config_space.add_hyperparameter(
        #     CS.UniformFloatHyperparameter(
        #         "gc_weight", lower=0, upper=1, default_value=1
        #     )
        # )


        return config_space

    @staticmethod
    def _fill_config(config):
        config["conv_size1"] = 1 + 2 * config["conv_radius1"]
        if config["conv_radius1"] == 0:
            config["conv_size1"] = 0
        del config["conv_radius1"]

        config["conv_size2"] = 1 + 2 * config["conv_radius2"]
        if config["conv_radius2"] == 0:
            config["conv_size2"] = 0
        del config["conv_radius2"]

        if config["conv_size1"] != 0:
            min_state_radius = config["conv_size1"] + config["conv_size1"] - 1
            max_state_radius = 32  # FR changed max state radius from 32 to 64, ICLR: LEARNA (32), Meta-LEARNA (29)
            config["state_radius"] = int(
                min_state_radius
                + (max_state_radius - min_state_radius) * config["state_radius_relative"]
            )
        else:
            min_state_radius = config["conv_size2"] + config["conv_size2"] - 1
            max_state_radius = 32  # FR changed max state radius from 32 to 64, ICLR: LEARNA (32), Meta-LEARNA (29)
            config["state_radius"] = int(
                min_state_radius
                + (max_state_radius - min_state_radius) * config["state_radius_relative"]
            )
        del config["state_radius_relative"]

        # config["desired_gc"] = np.random.choice([.1, .2, .3, .4, .5, .6, .7, .8, .9])
        config["restart_timeout"] = None
        config["local_design"] = True
        # config["reward_function"] = 'structure_only'
        # config["predict_pairs"] = 1

        return config



import os

os.environ["OMP_NUM_THREADS"] = "1"

import logging

logging.basicConfig(level=logging.ERROR)

import argparse
import pickle
import time
import json

import hpbandster.core.nameserver as hpns
import hpbandster.core.result as hpres

from hpbandster.core.base_iteration import  Datum
from hpbandster.core.result import Result
from hpbandster.optimizers import BOHB as BOHB

from src.optimization.learna_worker import LearnaWorker
from src.optimization.meta_learna_worker import MetaLearnaWorker
from src.optimization.meta_learna_adapt_worker import MetaLearnaAdaptWorker


def load_valid_results(directory):
    data = {}
    time_ref = float('inf')
    budget_set = set()

    with open(os.path.join(directory, 'configs.json')) as fh:
        for line in fh:

            line = json.loads(line)

            if len(line) == 3:
                config_id, config, config_info = line
            if len(line) == 2:
                config_id, config, = line
                config_info = 'N/A'
            data[tuple(config_id)] = Datum(config=config, config_info=config_info)


    exceptions = []
    with open(os.path.join(directory, 'results.json')) as fh:
        for line in fh:
            config_id, budget,time_stamps, result, exception = json.loads(line)

            if exception:
                print("found exception: \n" + f"{exception}" + f"in config: {config_id}")
                exceptions.append(config_id)
                continue


            id = tuple(config_id)

            data[id].time_stamps[budget] = time_stamps
            data[id].results[budget] = result
            data[id].exceptions[budget] = exception

            budget_set.add(budget)
            time_ref = min(time_ref, time_stamps['submitted'])

    for id in exceptions:
        id = tuple(id)
        del data[id]


    # infer the hyperband configuration from the data
    budget_list = sorted(list(budget_set))

    HB_config = {
                        'eta'        : None if len(budget_list) < 2 else budget_list[1]/budget_list[0],
                        'min_budget' : min(budget_set),
                        'max_budget' : max(budget_set),
                        'budgets'    : budget_list,
                        'max_SH_iter': len(budget_set),
                        'time_ref'   : time_ref
                }
    return(Result([data], HB_config))


parser = argparse.ArgumentParser(
    description="Optimization pipeline for Learning to Design RNA"
)
parser.add_argument(
    "--min_budget",
    type=float,
    help="Minimum budget used during the optimization.",
    default=10,
)
parser.add_argument(
    "--max_budget",
    type=float,
    help="Maximum budget used during the optimization.",
    default=10,
)
parser.add_argument(
    "--n_iterations",
    type=int,
    help="Number of iterations performed by the optimizer",
    default=16,
)

parser.add_argument(
    "--n_cores", type=int, help="Number of workers to run in parallel.", default=10
)
parser.add_argument(
    "--worker", help="Flag to turn this into a worker process", action="store_true"
)

parser.add_argument(
    "--run_id",
    type=str,
    help="A unique run id for this optimization run."
    "An easy option is to use the job id of the clusters scheduler.",
)
parser.add_argument("--data_dir", type=str, help="path where the datasets are stored")
parser.add_argument(
    "--nic_name", type=str, help="Which network interface to use for communication."
)
parser.add_argument(
    "--shared_directory",
    type=str,
    help="A directory that is accessible for all processes, e.g. a NFS share.",
)

parser.add_argument("--mode", choices=["learna", "meta_learna", "meta_learna_adapt"], default="learna")
# parser.add_argument(
#     "--load_existing", help="Flag to load exsting data", action="store_true"
# )
# parser.add_argument(
#     "--previous_run_dir",
#     type=str,
#     help="The directory of a previous run to load the data from.",
# )


# args=parser.parse_args("--run_id test --nic_name lo --shared_directory /tmp --n_cores 4 --data_dir src/data --mode L2DesignRNA".split())
args = parser.parse_args()

os.makedirs(args.shared_directory, exist_ok=True)

if args.mode == "learna":
    worker_cls = LearnaWorker
    worker_args = dict(
        data_dir=args.data_dir, num_cores=args.n_cores, train_sequences=range(1, 101)
    )

if args.mode == "meta_learna":
    worker_cls = MetaLearnaWorker
    worker_args = dict(
        data_dir=args.data_dir,
        num_cores=args.n_cores,
        train_sequences=range(1, 100),
        validation_timeout=1,
    )

if args.mode == "meta_learna_adapt":
    worker_cls = MetaLearnaAdaptWorker
    worker_args = dict(
        data_dir=args.data_dir,
        num_cores=args.n_cores,
        train_sequences=range(1, 100000),
        validation_timeout=60,
    )


# Every process has to lookup the hostname
host = hpns.nic_name_to_host(args.nic_name)

if args.worker:
    time.sleep(5)  # short artificial delay to make sure the nameserver is already running
    w = worker_cls(**worker_args, run_id=args.run_id, host=host)
    w.load_nameserver_credentials(working_directory=args.shared_directory)
    w.run(background=False)
    exit(0)
result_logger = hpres.json_result_logger(args.shared_directory, overwrite=True)


# Start a nameserver:
# We now start the nameserver with the host name from above and a random open port
NS = hpns.NameServer(
    run_id=args.run_id, host=host, port=0, working_directory=args.shared_directory
)
ns_host, ns_port = NS.start()

# Most optimizers are so computationally inexpensive that we can affort to run a
# worker in parallel to it. Note that this one has to run in the background to
# not plock!
w = worker_cls(
    **worker_args,
    run_id=args.run_id,
    host=host,
    nameserver=ns_host,
    nameserver_port=ns_port
)
w.run(background=True)

print(worker_cls.get_configspace())

print("############# LOADING PREVIOUS RUN")
# previous_run = load_valid_results(args.previous_run_dir)

# Run an optimizer
# We now have to specify the host, and the nameserver information
bohb = BOHB(
    configspace=worker_cls.get_configspace(),
    run_id=args.run_id,
    host=host,
    nameserver=ns_host,
    nameserver_port=ns_port,
    min_budget=args.min_budget,
    max_budget=args.max_budget,
    result_logger=result_logger,
    ping_interval=600,
    working_directory=args.shared_directory,
    # previous_result = previous_run,
)
res = bohb.run(n_iterations=args.n_iterations, min_n_workers=1)


# In a cluster environment, you usually want to store the results for later analysis.
# One option is to simply pickle the Result object
with open(os.path.join(args.shared_directory, "results.pkl"), "wb") as fh:
    pickle.dump(res, fh)


# Step 4: Shutdown
# After the optimizer run, we must shutdown the master and the nameserver.
bohb.shutdown(shutdown_workers=True)
NS.shutdown()



"""
    Testsuite for RNA-Design environment.
"""

import pytest
import numpy as np
import numpy.testing as nt

from distance import hamming
from itertools import product

from learna_tools.learna.environment import _string_difference_indices
from learna_tools.learna.environment import RnaDesignEnvironmentConfig
from learna_tools.learna.environment import _encode_dot_bracket
from learna_tools.learna.environment import _encode_pairing
from learna_tools.learna.environment import _Target
from learna_tools.learna.environment import _Design
from learna_tools.learna.environment import RnaDesignEnvironment

from RNA import fold


def test_string_difference_indices():
    # Test general behaviour
    s1 = "...(((..))).."
    s2 = "...((....)).."

    assert [5, 8] == _string_difference_indices(s1, s2)

    # Test equal strings
    s1 = "....."
    s2 = "....."

    assert [] == _string_difference_indices(s1, s2)

    # Test empty input s1
    s1 = ""
    s2 = "...().."

    assert [] == _string_difference_indices(s1, s2)

    # Test empty input s2
    s1 = "....."
    s2 = ""

    nt.assert_raises(
        IndexError, _string_difference_indices, s1, s2
    )  # TODO: Handle IndexError


def test_encode_dot_bracket():
    secondary = "..((..))."
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=0
    )

    # Assert general behaviour
    assert [int(site) for site in "001100110"] == _encode_dot_bracket(
        secondary, environment_config
    )

    # Test padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=1
    )
    assert [int(site) for site in "00011001100"] == _encode_dot_bracket(
        secondary, environment_config
    )

    # Test embedding encoding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=0
    )
    assert [int(site) for site in "001100220"] == _encode_dot_bracket(
        secondary, environment_config
    )

    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=1
    )
    assert [int(site) for site in "30011002203"] == _encode_dot_bracket(
        secondary, environment_config
    )

    # Test convolution encoding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=0
    )
    assert [[int(site)] for site in "001100110"] == _encode_dot_bracket(
        secondary, environment_config
    )

    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=1
    )
    assert [[int(site)] for site in "00011001100"] == _encode_dot_bracket(
        secondary, environment_config
    )

    # Ignore convolution encoding if embedding is used
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=0
    )
    assert [int(site) for site in "001100220"] == _encode_dot_bracket(
        secondary, environment_config
    )

    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=1
    )
    assert [int(site) for site in "30011002203"] == _encode_dot_bracket(
        secondary, environment_config
    )


def test_encode_pairing():
    secondary = "..((..))."

    # Test general behaviour
    assert [None, None, 7, 6, None, None, 3, 2, None] == _encode_pairing(secondary)

    # Test no pairs in sequence
    secondary = "....."
    assert [None for site in secondary] == _encode_pairing(secondary)

    # Test sequence of pairs only
    secondary = "((()))"
    assert [5, 4, 3, 2, 1, 0] == _encode_pairing(secondary)

    # Test empty input
    secondary = ""
    assert [] == _encode_pairing(secondary)


def test_Target():
    dot_bracket = "..((..))."

    # No conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=0
    )

    target = _Target(dot_bracket, environment_config)
    assert 1 == target.id
    assert "..((..))." == target.dot_bracket
    assert [int(site) for site in "001100110"] == target.padded_encoding

    [
        nt.assert_equal(site, target.get_paired_site(index))
        for index, site in enumerate([None, None, 7, 6, None, None, 3, 2, None])
    ]

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=1
    )

    target = _Target(dot_bracket, environment_config)
    assert 2 == target.id
    assert "..((..))." == target.dot_bracket
    assert [int(site) for site in "00011001100"] == target.padded_encoding

    [
        nt.assert_equal(site, target.get_paired_site(index))
        for index, site in enumerate([None, None, 7, 6, None, None, 3, 2, None])
    ]

    # No conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=0
    )

    target = _Target(dot_bracket, environment_config)
    assert 3 == target.id
    assert "..((..))." == target.dot_bracket
    assert [int(site) for site in "001100220"] == target.padded_encoding

    [
        nt.assert_equal(site, target.get_paired_site(index))
        for index, site in enumerate([None, None, 7, 6, None, None, 3, 2, None])
    ]

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=1
    )

    target = _Target(dot_bracket, environment_config)
    assert 4 == target.id
    assert "..((..))." == target.dot_bracket
    assert [int(site) for site in "30011002203"] == target.padded_encoding

    [
        nt.assert_equal(site, target.get_paired_site(index))
        for index, site in enumerate([None, None, 7, 6, None, None, 3, 2, None])
    ]

    # Conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=0
    )

    target = _Target(dot_bracket, environment_config)
    assert 5 == target.id
    assert "..((..))." == target.dot_bracket
    assert [int(site) for site in "001100220"] == target.padded_encoding

    [
        nt.assert_equal(site, target.get_paired_site(index))
        for index, site in enumerate([None, None, 7, 6, None, None, 3, 2, None])
    ]

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=1
    )

    target = _Target(dot_bracket, environment_config)
    assert 6 == target.id
    assert "..((..))." == target.dot_bracket
    assert [int(site) for site in "30011002203"] == target.padded_encoding

    [
        nt.assert_equal(site, target.get_paired_site(index))
        for index, site in enumerate([None, None, 7, 6, None, None, 3, 2, None])
    ]

    # Conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=0
    )

    target = _Target(dot_bracket, environment_config)
    assert 7 == target.id
    assert "..((..))." == target.dot_bracket
    assert [[int(site)] for site in "001100110"] == target.padded_encoding

    [
        nt.assert_equal(site, target.get_paired_site(index))
        for index, site in enumerate([None, None, 7, 6, None, None, 3, 2, None])
    ]

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=1
    )

    target = _Target(dot_bracket, environment_config)
    assert 8 == target.id
    assert "..((..))." == target.dot_bracket
    assert [[int(site)] for site in "00011001100"] == target.padded_encoding

    [
        nt.assert_equal(site, target.get_paired_site(index))
        for index, site in enumerate([None, None, 7, 6, None, None, 3, 2, None])
    ]


def test_Design():
    # Test Initialization
    with nt.assert_raises(TypeError):
        design = _Design()  # TODO: Handle TypeError caused by NoneType initialization
    dot_bracket = "..((..))."

    # No conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=0
    )

    target = _Target(dot_bracket, environment_config)
    design = _Design(length=len(target))

    with nt.assert_raises(TypeError):
        design.primary  # TODO: Handle TypeError caused by NoneType init

    assert design.first_unassigned_site == 0

    # Test nucleotide assignment
    for i in range(len(dot_bracket)):
        if dot_bracket[i] == ")":
            continue
        action = 0
        site = design.first_unassigned_site
        design.assign_sites(action, site, paired_site=target.get_paired_site(site))

    assert "GGGGGGCCG" == design.primary

    # Test Local Improvement procedure
    mutations = "AGCU"

    for mutation in mutations:
        site = [1]
        mutated = design.get_mutated(mutation, site)
        assert f"G{mutation}GGGGCCG" == mutated.primary

        site = [2]
        mutated = design.get_mutated(mutation, site)
        assert f"GG{mutation}GGGCCG" == mutated.primary

        site = [3]
        mutated = design.get_mutated(mutation, site)
        assert f"GGG{mutation}GGCCG" == mutated.primary

        site = [4]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGG{mutation}GCCG" == mutated.primary

        site = [5]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGG{mutation}CCG" == mutated.primary

        site = [6]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGG{mutation}CG" == mutated.primary

        site = [7]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGC{mutation}G" == mutated.primary

        site = [8]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGCC{mutation}" == mutated.primary

        with nt.assert_raises(IndexError):
            site = [9]
            mutated = design.get_mutated(mutation, site)

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=1
    )

    target = _Target(dot_bracket, environment_config)
    design = _Design(length=len(target))

    with nt.assert_raises(TypeError):
        design.primary  # TODO: Handle TypeError caused by NoneType init

    assert design.first_unassigned_site == 0

    # Test nucleotide assignment
    for i in range(len(dot_bracket)):
        if dot_bracket[i] == ")":
            continue
        action = 0
        site = design.first_unassigned_site
        design.assign_sites(action, site, paired_site=target.get_paired_site(site))

    assert "GGGGGGCCG" == design.primary

    # Test Local Improvement procedure
    mutations = "AGCU"

    for mutation in mutations:
        site = [1]
        mutated = design.get_mutated(mutation, site)
        assert f"G{mutation}GGGGCCG" == mutated.primary

        site = [2]
        mutated = design.get_mutated(mutation, site)
        assert f"GG{mutation}GGGCCG" == mutated.primary

        site = [3]
        mutated = design.get_mutated(mutation, site)
        assert f"GGG{mutation}GGCCG" == mutated.primary

        site = [4]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGG{mutation}GCCG" == mutated.primary

        site = [5]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGG{mutation}CCG" == mutated.primary

        site = [6]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGG{mutation}CG" == mutated.primary

        site = [7]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGC{mutation}G" == mutated.primary

        site = [8]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGCC{mutation}" == mutated.primary

        with nt.assert_raises(IndexError):
            site = [9]
            mutated = design.get_mutated(mutation, site)

    # No conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=0
    )

    target = _Target(dot_bracket, environment_config)
    design = _Design(length=len(target))

    with nt.assert_raises(TypeError):
        design.primary  # TODO: Handle TypeError caused by NoneType init

    assert design.first_unassigned_site == 0

    # Test nucleotide assignment
    for i in range(len(dot_bracket)):
        if dot_bracket[i] == ")":
            continue
        action = 0
        site = design.first_unassigned_site
        design.assign_sites(action, site, paired_site=target.get_paired_site(site))

    assert "GGGGGGCCG" == design.primary

    # Test Local Improvement procedure
    mutations = "AGCU"

    for mutation in mutations:
        site = [1]
        mutated = design.get_mutated(mutation, site)
        assert f"G{mutation}GGGGCCG" == mutated.primary

        site = [2]
        mutated = design.get_mutated(mutation, site)
        assert f"GG{mutation}GGGCCG" == mutated.primary

        site = [3]
        mutated = design.get_mutated(mutation, site)
        assert f"GGG{mutation}GGCCG" == mutated.primary

        site = [4]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGG{mutation}GCCG" == mutated.primary

        site = [5]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGG{mutation}CCG" == mutated.primary

        site = [6]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGG{mutation}CG" == mutated.primary

        site = [7]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGC{mutation}G" == mutated.primary

        site = [8]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGCC{mutation}" == mutated.primary

        with nt.assert_raises(IndexError):
            site = [9]
            mutated = design.get_mutated(mutation, site)

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=1
    )

    target = _Target(dot_bracket, environment_config)
    design = _Design(length=len(target))

    with nt.assert_raises(TypeError):
        design.primary  # TODO: Handle TypeError caused by NoneType init

    assert design.first_unassigned_site == 0

    # Test nucleotide assignment
    for i in range(len(dot_bracket)):
        if dot_bracket[i] == ")":
            continue
        action = 0
        site = design.first_unassigned_site
        design.assign_sites(action, site, paired_site=target.get_paired_site(site))

    assert "GGGGGGCCG" == design.primary

    # Test Local Improvement procedure
    mutations = "AGCU"

    for mutation in mutations:
        site = [1]
        mutated = design.get_mutated(mutation, site)
        assert f"G{mutation}GGGGCCG" == mutated.primary

        site = [2]
        mutated = design.get_mutated(mutation, site)
        assert f"GG{mutation}GGGCCG" == mutated.primary

        site = [3]
        mutated = design.get_mutated(mutation, site)
        assert f"GGG{mutation}GGCCG" == mutated.primary

        site = [4]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGG{mutation}GCCG" == mutated.primary

        site = [5]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGG{mutation}CCG" == mutated.primary

        site = [6]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGG{mutation}CG" == mutated.primary

        site = [7]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGC{mutation}G" == mutated.primary

        site = [8]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGCC{mutation}" == mutated.primary

        with nt.assert_raises(IndexError):
            site = [9]
            mutated = design.get_mutated(mutation, site)

    # Conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=0
    )

    target = _Target(dot_bracket, environment_config)
    design = _Design(length=len(target))

    with nt.assert_raises(TypeError):
        design.primary  # TODO: Handle TypeError caused by NoneType init

    assert design.first_unassigned_site == 0

    # Test nucleotide assignment
    for i in range(len(dot_bracket)):
        if dot_bracket[i] == ")":
            continue
        action = 0
        site = design.first_unassigned_site
        design.assign_sites(action, site, paired_site=target.get_paired_site(site))

    assert "GGGGGGCCG" == design.primary

    # Test Local Improvement procedure
    mutations = "AGCU"

    for mutation in mutations:
        site = [1]
        mutated = design.get_mutated(mutation, site)
        assert f"G{mutation}GGGGCCG" == mutated.primary

        site = [2]
        mutated = design.get_mutated(mutation, site)
        assert f"GG{mutation}GGGCCG" == mutated.primary

        site = [3]
        mutated = design.get_mutated(mutation, site)
        assert f"GGG{mutation}GGCCG" == mutated.primary

        site = [4]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGG{mutation}GCCG" == mutated.primary

        site = [5]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGG{mutation}CCG" == mutated.primary

        site = [6]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGG{mutation}CG" == mutated.primary

        site = [7]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGC{mutation}G" == mutated.primary

        site = [8]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGCC{mutation}" == mutated.primary

        with nt.assert_raises(IndexError):
            site = [9]
            mutated = design.get_mutated(mutation, site)

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=1
    )

    target = _Target(dot_bracket, environment_config)
    design = _Design(length=len(target))

    with nt.assert_raises(TypeError):
        design.primary  # TODO: Handle TypeError caused by NoneType init

    assert design.first_unassigned_site == 0

    # Test nucleotide assignment
    for i in range(len(dot_bracket)):
        if dot_bracket[i] == ")":
            continue
        action = 0
        site = design.first_unassigned_site
        design.assign_sites(action, site, paired_site=target.get_paired_site(site))

    assert "GGGGGGCCG" == design.primary

    # Test Local Improvement procedure
    mutations = "AGCU"

    for mutation in mutations:
        site = [1]
        mutated = design.get_mutated(mutation, site)
        assert f"G{mutation}GGGGCCG" == mutated.primary

        site = [2]
        mutated = design.get_mutated(mutation, site)
        assert f"GG{mutation}GGGCCG" == mutated.primary

        site = [3]
        mutated = design.get_mutated(mutation, site)
        assert f"GGG{mutation}GGCCG" == mutated.primary

        site = [4]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGG{mutation}GCCG" == mutated.primary

        site = [5]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGG{mutation}CCG" == mutated.primary

        site = [6]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGG{mutation}CG" == mutated.primary

        site = [7]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGC{mutation}G" == mutated.primary

        site = [8]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGCC{mutation}" == mutated.primary

        with nt.assert_raises(IndexError):
            site = [9]
            mutated = design.get_mutated(mutation, site)

    # Conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=0
    )

    target = _Target(dot_bracket, environment_config)
    design = _Design(length=len(target))

    with nt.assert_raises(TypeError):
        design.primary  # TODO: Handle TypeError caused by NoneType init

    assert design.first_unassigned_site == 0

    # Test nucleotide assignment
    for i in range(len(dot_bracket)):
        if dot_bracket[i] == ")":
            continue
        action = 0
        site = design.first_unassigned_site
        design.assign_sites(action, site, paired_site=target.get_paired_site(site))

    assert "GGGGGGCCG" == design.primary

    # Test Local Improvement procedure
    mutations = "AGCU"

    for mutation in mutations:
        site = [1]
        mutated = design.get_mutated(mutation, site)
        assert f"G{mutation}GGGGCCG" == mutated.primary

        site = [2]
        mutated = design.get_mutated(mutation, site)
        assert f"GG{mutation}GGGCCG" == mutated.primary

        site = [3]
        mutated = design.get_mutated(mutation, site)
        assert f"GGG{mutation}GGCCG" == mutated.primary

        site = [4]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGG{mutation}GCCG" == mutated.primary

        site = [5]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGG{mutation}CCG" == mutated.primary

        site = [6]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGG{mutation}CG" == mutated.primary

        site = [7]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGC{mutation}G" == mutated.primary

        site = [8]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGCC{mutation}" == mutated.primary

        with nt.assert_raises(IndexError):
            site = [9]
            mutated = design.get_mutated(mutation, site)

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=1
    )

    target = _Target(dot_bracket, environment_config)
    design = _Design(length=len(target))

    with nt.assert_raises(TypeError):
        design.primary  # TODO: Handle TypeError caused by NoneType init

    assert design.first_unassigned_site == 0

    # Test nucleotide assignment
    for i in range(len(dot_bracket)):
        if dot_bracket[i] == ")":
            continue
        action = 0
        site = design.first_unassigned_site
        design.assign_sites(action, site, paired_site=target.get_paired_site(site))

    assert "GGGGGGCCG" == design.primary

    # Test Local Improvement procedure
    mutations = "AGCU"

    for mutation in mutations:
        site = [1]
        mutated = design.get_mutated(mutation, site)
        assert f"G{mutation}GGGGCCG" == mutated.primary

        site = [2]
        mutated = design.get_mutated(mutation, site)
        assert f"GG{mutation}GGGCCG" == mutated.primary

        site = [3]
        mutated = design.get_mutated(mutation, site)
        assert f"GGG{mutation}GGCCG" == mutated.primary

        site = [4]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGG{mutation}GCCG" == mutated.primary

        site = [5]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGG{mutation}CCG" == mutated.primary

        site = [6]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGG{mutation}CG" == mutated.primary

        site = [7]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGC{mutation}G" == mutated.primary

        site = [8]
        mutated = design.get_mutated(mutation, site)
        assert f"GGGGGGCC{mutation}" == mutated.primary

        with nt.assert_raises(IndexError):
            site = [9]
            mutated = design.get_mutated(mutation, site)


def test_RnaDesignEnvironment_reset():
    dot_brackets = ["..((..))."]

    # No conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    assert None == environment.target
    assert None == environment.design

    assert [0] == environment.reset()
    assert "..((..))." == environment.target.dot_bracket
    assert [int(site) for site in "001100110"] == environment.target.padded_encoding

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    assert None == environment.target
    assert None == environment.design

    assert [0, 0, 0] == environment.reset()
    assert "..((..))." == environment.target.dot_bracket
    assert [int(site) for site in "00011001100"] == environment.target.padded_encoding

    # No conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    assert None == environment.target
    assert None == environment.design

    assert [0] == environment.reset()
    assert "..((..))." == environment.target.dot_bracket
    assert [int(site) for site in "001100220"] == environment.target.padded_encoding

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    assert None == environment.target
    assert None == environment.design

    assert [3, 0, 0] == environment.reset()
    assert "..((..))." == environment.target.dot_bracket
    assert [int(site) for site in "30011002203"] == environment.target.padded_encoding

    # Conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    assert None == environment.target
    assert None == environment.design

    assert [[0]] == environment.reset()
    assert "..((..))." == environment.target.dot_bracket
    assert [[int(site)] for site in "001100110"] == environment.target.padded_encoding

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    assert None == environment.target
    assert None == environment.design

    assert [[0], [0], [0]] == environment.reset()
    assert "..((..))." == environment.target.dot_bracket
    assert [[int(site)] for site in "00011001100"] == environment.target.padded_encoding

    # Conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    assert None == environment.target
    assert None == environment.design

    assert [0] == environment.reset()
    assert "..((..))." == environment.target.dot_bracket
    assert [int(site) for site in "001100220"] == environment.target.padded_encoding

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    assert None == environment.target
    assert None == environment.design

    assert [3, 0, 0] == environment.reset()
    assert "..((..))." == environment.target.dot_bracket
    assert [int(site) for site in "30011002203"] == environment.target.padded_encoding


def test_RnaDesignEnvironment_apply_action():
    primary_list = list("GGGGGGCCG")
    dot_brackets = ["..((..))."]

    # No conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    i = 0

    while environment.design.first_unassigned_site is not None:
        action = 0
        environment._apply_action(action)
        assert environment.design._primary_list[:i] == primary_list[:i]
        i += 1
    assert "GGGGGGCCG" == environment.design.primary

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    i = 0

    while environment.design.first_unassigned_site is not None:
        action = 0
        environment._apply_action(action)
        assert environment.design._primary_list[:i] == primary_list[:i]
        i += 1
    assert "GGGGGGCCG" == environment.design.primary

    # No conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    i = 0

    while environment.design.first_unassigned_site is not None:
        action = 0
        environment._apply_action(action)
        assert environment.design._primary_list[:i] == primary_list[:i]
        i += 1
    assert "GGGGGGCCG" == environment.design.primary

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    i = 0

    while environment.design.first_unassigned_site is not None:
        action = 0
        environment._apply_action(action)
        assert environment.design._primary_list[:i] == primary_list[:i]
        i += 1
    assert "GGGGGGCCG" == environment.design.primary

    # Conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    i = 0

    while environment.design.first_unassigned_site is not None:
        action = 0
        environment._apply_action(action)
        assert environment.design._primary_list[:i] == primary_list[:i]
        i += 1
    assert "GGGGGGCCG" == environment.design.primary

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    i = 0

    while environment.design.first_unassigned_site is not None:
        action = 0
        environment._apply_action(action)
        assert environment.design._primary_list[:i] == primary_list[:i]
        i += 1
    assert "GGGGGGCCG" == environment.design.primary

    # Conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    i = 0

    while environment.design.first_unassigned_site is not None:
        action = 0
        environment._apply_action(action)
        assert environment.design._primary_list[:i] == primary_list[:i]
        i += 1
    assert "GGGGGGCCG" == environment.design.primary

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    i = 0

    while environment.design.first_unassigned_site is not None:
        action = 0
        environment._apply_action(action)
        assert environment.design._primary_list[:i] == primary_list[:i]
        i += 1
    assert "GGGGGGCCG" == environment.design.primary


def test_RnaDesignEnvironment_get_state():
    dot_brackets = ["..((..))."]

    # No conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    states = [[state] for state in environment.target.padded_encoding]

    while environment.design.first_unassigned_site is not None:
        assert (
            states[environment.design.first_unassigned_site] == environment._get_state()
        )
        environment._apply_action(0)

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    states = [
        [
            environment.target.padded_encoding[i - 1],
            environment.target.padded_encoding[i],
            environment.target.padded_encoding[i + 1],
        ]
        for i in range(len(environment.target.padded_encoding) - 1)
        if i > 0
    ]

    while environment.design.first_unassigned_site is not None:
        assert (
            states[environment.design.first_unassigned_site] == environment._get_state()
        )
        environment._apply_action(0)

    # No conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    states = [[state] for state in environment.target.padded_encoding]

    while environment.design.first_unassigned_site is not None:
        assert (
            states[environment.design.first_unassigned_site] == environment._get_state()
        )
        environment._apply_action(0)

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    states = [
        [
            environment.target.padded_encoding[i - 1],
            environment.target.padded_encoding[i],
            environment.target.padded_encoding[i + 1],
        ]
        for i in range(len(environment.target.padded_encoding) - 1)
        if i > 0
    ]

    while environment.design.first_unassigned_site is not None:
        assert (
            states[environment.design.first_unassigned_site] == environment._get_state()
        )
        environment._apply_action(0)

    # Conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    states = [[state] for state in environment.target.padded_encoding]

    while environment.design.first_unassigned_site is not None:
        assert (
            states[environment.design.first_unassigned_site] == environment._get_state()
        )
        environment._apply_action(0)

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    states = [
        [
            environment.target.padded_encoding[i - 1],
            environment.target.padded_encoding[i],
            environment.target.padded_encoding[i + 1],
        ]
        for i in range(len(environment.target.padded_encoding) - 1)
        if i > 0
    ]

    while environment.design.first_unassigned_site is not None:
        assert (
            states[environment.design.first_unassigned_site] == environment._get_state()
        )
        environment._apply_action(0)

    # Conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    states = [[state] for state in environment.target.padded_encoding]

    while environment.design.first_unassigned_site is not None:
        assert (
            states[environment.design.first_unassigned_site] == environment._get_state()
        )
        environment._apply_action(0)

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    states = [
        [
            environment.target.padded_encoding[i - 1],
            environment.target.padded_encoding[i],
            environment.target.padded_encoding[i + 1],
        ]
        for i in range(len(environment.target.padded_encoding) - 1)
        if i > 0
    ]

    while environment.design.first_unassigned_site is not None:
        assert (
            states[environment.design.first_unassigned_site] == environment._get_state()
        )
        environment._apply_action(0)


# TODO: Find example with final distance > 0 to test for distance list
def test_RnaDesignEnvironment_local_improvement():
    dot_brackets = ["((....))"]

    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    # Assign nucleotides such that design.primary folds correctly
    # Possible solution: GCGAUAGC
    environment._apply_action(0)
    environment._apply_action(1)
    environment._apply_action(0)
    environment._apply_action(1)
    environment._apply_action(2)
    environment._apply_action(1)

    assert "GCGAUAGC" == environment.design.primary

    # NOTE: Hamming distances may differ if other folding algorithm or new version is used
    assert 0 == hamming(
        environment.target.dot_bracket, fold(environment.design.primary)[0]
    )

    # Mutate primary sequence before local improvement step
    mutated = environment.design.get_mutated("A", [0])
    assert "ACGAUAGC" == mutated.primary
    assert "........" == fold(mutated.primary)[0]
    hamming_distance = hamming(environment.target.dot_bracket, fold(mutated.primary)[0])
    assert 0 < hamming_distance
    assert 4 == hamming_distance
    assert 0 == environment._local_improvement(fold(mutated.primary)[0])

    mutated = environment.design.get_mutated("AA", [0, 1])
    assert "AAGAUAGC" == mutated.primary
    assert "........" == fold(mutated.primary)[0]
    hamming_distance = hamming(environment.target.dot_bracket, fold(mutated.primary)[0])
    assert 0 < hamming_distance
    assert 4 == hamming_distance
    assert 0 == environment._local_improvement(fold(mutated.primary)[0])

    mutated = environment.design.get_mutated("AAAAAAAA", [0, 1, 2, 3, 4, 5, 6, 7])
    assert "AAAAAAAA" == mutated.primary
    assert "........" == fold(mutated.primary)[0]
    hamming_distance = hamming(environment.target.dot_bracket, fold(mutated.primary)[0])
    assert 0 < hamming_distance
    assert 4 == hamming_distance
    assert 0 == environment._local_improvement(fold(mutated.primary)[0])


def test_RnaDesignEnvironment_get_reward():
    dot_brackets = ["(((....)))"]

    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)
    environment.reset()

    # Assign nucleotides such that design.primary folds correctly
    # Possible solution: CGCCUACGCG
    environment._apply_action(1)
    environment._apply_action(0)
    environment._apply_action(1)
    environment._apply_action(3)
    environment._apply_action(2)
    environment._apply_action(1)
    environment._apply_action(3)

    # Test general behaviour
    assert "CGCCUACGCG" == environment.design.primary
    assert 0 == environment._get_reward(False)
    assert 1.0 == environment._get_reward(True)

    # Mutate primary sequence to change secondary structure
    # NOTE: Hamming distance is larger than mutation_threshold (5), no local improvement is applied
    mutated = environment.design.get_mutated(
        tuple(list("AAAAAAAAAA")), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    )
    assert "AAAAAAAAAA" == mutated.primary
    assert ".........." == fold(mutated.primary)[0]
    hamming_distance = hamming(environment.target.dot_bracket, fold(mutated.primary)[0])
    assert 0 < hamming_distance
    assert 6 == hamming_distance

    environment.design = mutated
    assert 1.0 > environment._get_reward(True)

    hamming_distance = hamming(
        environment.target.dot_bracket, fold(environment.design.primary)[0]
    )
    normalized_hamming_distance = (
        1
        - (hamming_distance / len(environment.target))
        ** environment._env_config.reward_exponent
    )
    assert normalized_hamming_distance == environment._get_reward(True)

    # Mutate primary sequence to change secondary structure
    # NOTE: Hamming distance is smaller than mutation_threshold (5), local improvement is applied
    mutated = environment.design.get_mutated(
        tuple(list("GGCCUAUGCG")), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    )
    assert "GGCCUAUGCG" == mutated.primary
    assert ".((....))." == fold(mutated.primary)[0]
    hamming_distance = hamming(environment.target.dot_bracket, fold(mutated.primary)[0])
    assert 0 < hamming_distance
    assert 2 == hamming_distance

    environment.design = mutated
    assert 1.0 == environment._get_reward(True)


def test_RnaDesignEnvironment_execute():
    actions = [1, 0, 1, 3, 2, 1, 3]  # Actions correspond to valid solution
    dot_brackets = ["(((....)))"]

    # No conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [[state] for state in environment.target.padded_encoding][
        1:
    ]  # reset() already gets first state

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [
        [
            environment.target.padded_encoding[i - 1],
            environment.target.padded_encoding[i],
            environment.target.padded_encoding[i + 1],
        ]
        for i in range(len(environment.target.padded_encoding) - 1)
        if i > 0
    ][1:]

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # No conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [[state] for state in environment.target.padded_encoding][
        1:
    ]  # reset() already gets first state

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [
        [
            environment.target.padded_encoding[i - 1],
            environment.target.padded_encoding[i],
            environment.target.padded_encoding[i + 1],
        ]
        for i in range(len(environment.target.padded_encoding) - 1)
        if i > 0
    ][1:]

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # Conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [[state] for state in environment.target.padded_encoding][
        1:
    ]  # reset() already gets first state

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [
        [
            environment.target.padded_encoding[i - 1],
            environment.target.padded_encoding[i],
            environment.target.padded_encoding[i + 1],
        ]
        for i in range(len(environment.target.padded_encoding) - 1)
        if i > 0
    ][1:]

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # Conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [[state] for state in environment.target.padded_encoding][
        1:
    ]  # reset() already gets first state

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [
        [
            environment.target.padded_encoding[i - 1],
            environment.target.padded_encoding[i],
            environment.target.padded_encoding[i + 1],
        ]
        for i in range(len(environment.target.padded_encoding) - 1)
        if i > 0
    ][1:]

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    actions = [0, 1, 0, 1, 2, 1]  # Actions correspond to valid solution
    dot_brackets = ["((....))"]

    # No conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [[state] for state in environment.target.padded_encoding][
        1:
    ]  # reset() already gets first state

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=False, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [
        [
            environment.target.padded_encoding[i - 1],
            environment.target.padded_encoding[i],
            environment.target.padded_encoding[i + 1],
        ]
        for i in range(len(environment.target.padded_encoding) - 1)
        if i > 0
    ][1:]

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # No conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [[state] for state in environment.target.padded_encoding][
        1:
    ]  # reset() already gets first state

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=False, use_embedding=True, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [
        [
            environment.target.padded_encoding[i - 1],
            environment.target.padded_encoding[i],
            environment.target.padded_encoding[i + 1],
        ]
        for i in range(len(environment.target.padded_encoding) - 1)
        if i > 0
    ][1:]

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # Conv, no embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [[state] for state in environment.target.padded_encoding][
        1:
    ]  # reset() already gets first state

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=False, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [
        [
            environment.target.padded_encoding[i - 1],
            environment.target.padded_encoding[i],
            environment.target.padded_encoding[i + 1],
        ]
        for i in range(len(environment.target.padded_encoding) - 1)
        if i > 0
    ][1:]

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # Conv, embedding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=0
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [[state] for state in environment.target.padded_encoding][
        1:
    ]  # reset() already gets first state

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal

    # Include padding
    environment_config = RnaDesignEnvironmentConfig(
        use_conv=True, use_embedding=True, state_radius=1
    )

    environment = RnaDesignEnvironment(dot_brackets, environment_config)

    environment.reset()
    states = [
        [
            environment.target.padded_encoding[i - 1],
            environment.target.padded_encoding[i],
            environment.target.padded_encoding[i + 1],
        ]
        for i in range(len(environment.target.padded_encoding) - 1)
        if i > 0
    ][1:]

    for index, action in enumerate(actions):
        state, terminal, reward = environment.execute(action)
        if terminal:
            assert None == state
            assert 1.0 == reward
            break
        assert states[index] == state
        assert 0 == reward
        assert False == terminal



import numpy as np
from collections import defaultdict

def number_of_mismatches(m1, m2):
    """
    counts the number of differences of the two matrices m1 and m2
    """
    missmatches = 0
    for x in range(len(m1)):
        for y in range(len(m1[x])):
            if m1[x][y] == m2[x][y]:
                missmatches += 1

    return missmatches


def db2mat(db):
    """
    Get matrix from dot-bracket notation of structure.
    Does not support partial RNA design with missing structure regions (Ns).

    Input: dot-bracket notation string or list
    Returns: Binary matrix with 1 indicating pair, 0 indicating no pair
    """
    corresponding_brackets = {')': '(', ']': '[', '}': '{', '>': '<'}
    mat = np.zeros((len(db), len(db)))
    pair_stack = defaultdict(list)

    for i, sym in enumerate(db):
        if sym == '.':
            continue
        elif sym in ['(', '[', '{', '<']:
            pair_stack[sym].append(i)
        else:
            open_idx = pair_stack[corresponding_brackets[sym]].pop()
            mat[open_idx, i] = 1
            mat[i, open_idx] = 1
    return mat


def score_matrix(true_mat, pred_mat):
    """
    Score two matrices with common metrics: F1-score, recall, specificity,
    precision, Matthews-Correlation-Coeficient (MCC).
    Does not support partial RNA design with missing structure regions.

    Input: true matrix, predicted matrix.
    Returns: Dict with scores.
    """
    solved = np.all(np.equal(true_mat, pred_mat)).astype(np.int8)
    if solved == 1:
        f1_score = 1
        precision = 1
        recall = 1
        specificity = 1
        mcc = 1
    else:
        # tp = true-positives
        # fp = false-positives
        # tn = true-negatives
        # fn = false-negatives
        tp = np.logical_and(pred_mat, true_mat).sum()
        non_correct = (tp == 0).astype(int)
        tn = np.logical_and(np.logical_not(pred_mat), np.logical_not(true_mat)).sum()
        fp = pred_mat.sum() - tp
        fn = true_mat.sum() - tp

        recall = tp / (tp + fn + 1e-8)
        precision = tp / (tp + fp + 1e-8)
        specificity = tn / (tn + fp + 1e-8)
        f1_score = 2 * tp / (2 * tp + fp + fn)
        mcc = (tp * tn - fp * fn) / np.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn) + 1e-8)

    mat_metrics = {'f1_score': f1_score, 'solved': solved}
    mat_metrics['precision'] = precision
    mat_metrics['recall'] = recall
    mat_metrics['specificity'] = specificity
    mat_metrics['mcc'] = mcc

    return mat_metrics



if __name__ == '__main__':
    from matplotlib import pyplot as plt

    true_strucs = ['....((((....))))....', '...(([[[))..]]]...', '..(())...']
    pred_strucs = ['...(((((....)))))...', '....(((((...))))).', '..(())...']
    solved = [False, False, True]

    for i, (t, p) in enumerate(zip(true_strucs, pred_strucs)):
        true = db2mat(t)
        pred = db2mat(p)
        scores = score_matrix(true, pred)
        assert scores['solved'] == solved[i]

        fix, axs = plt.subplots(1, 2)
        axs[0].imshow(true)
        axs[1].imshow(pred)
        plt.show()



from pathlib import Path

datasets = ['rfam_local_test', 'rfam_local_validation', 'rfam_local_train', 'rfam_local_min_400_max_1000_test', 'rfam_local_min_1000_test', 'rfam_local_short_train', 'rfam_local_long_train', 'rfam_learn_local_min_500_test', 'rfam_learn_local_min_100_max_500_test']

def get_dataset(data_dir, dataset):
    return Path(data_dir, dataset).glob('*.rna')

def validate(datasets):
    print('Validating all datasets')
    valid = [dataset.intersection(others) for dataset, others in zip(datasets.values(), datasets.values()) if not others == dataset]
    # if not all(valid
    validation_result = 'There are no intersections between datasets' if not valid else f"Datasets share targets: \n {valid}"
    print(validation_result)


def get_datasets(data_dir):
    print('reading all datasets')
    data = {}
    for dataset in datasets:
        dataset_targets = set([target.read_text().split('\t')[3] for target in get_dataset(data_dir, dataset)])
        data[dataset] = dataset_targets
    return data

def check(data_dir):
    for dataset in datasets:
        for index, target in enumerate(get_dataset(data_dir, dataset)):
            try:
                target.read_text().split('\t')[3]
            except:
                print(dataset, index)
                print(target.read_text().split('\t'))

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data_dir", type=str, default='data', help="The data dir"
    )

    args = parser.parse_args()

    validate(get_datasets(args.data_dir))
    # check(args.data_dir)



import pandas as pd
import numpy as np
from pathlib import Path

def split_local_data_to_single_files(data_dir, extension='.rna'):
    target_file_paths = Path(data_dir, dataset).glob('*.local')
    for target_file_path in target_file_paths:
        [Path(data_dir, f"{index+1}{extension}").resolve().write_text(target.rstrip('\n')) for index, target in enumerate(target_file_path.read_text().rstrip('\n').split('\n'))]

def split_via_dataframe(data_dir, dataset, in_extension='.interim', out_extension='.rna'):
    path = Path(data_dir, dataset, f"{str(dataset).split('_')[-1]}{in_extension}").resolve()
    print('Read interim data')
    df = pd.read_csv(path, sep='\t')
    print('Write single files')
    # dot_brackets = []
    # sequence_constraints = []
    # for target in df['local_random']:
    #     dot_brackets.append(target.upper().replace('A', 'N').replace('G', 'N').replace('C', 'N').replace('U', 'N'))
    #     sequence_constraints.append(target.upper().replace('.', 'N').replace('(', 'N').replace(')', 'N'))
    # [pd.DataFrame([(current_index +1, row[0], row[1], row[2], row[3], row[4], row[5])]).to_csv(path_or_buf=Path(data_dir, dataset, f"{current_index+1}{out_extension}"), sep='\t', index=False) for current_index, row in enumerate(zip(df['structure'], df['sequence'], df['local_random'], df['local_motif'], df['gc_content'], df['mfe']))]
    [Path(data_dir, dataset, f"{index+1}{out_extension}").resolve().write_text(f"{index+1}\t{row[0]}\t{row[1]}\t{row[2]}\t{row[3]}\t{row[4]}\t{row[5]}") for index, row in enumerate(zip(df['structure'], df['sequence'], df['local_random'], df['local_motif'], df['gc_content'], df['mfe']))]
    # [Path(data_dir, dataset, f"{index+1}{out_extension}").resolve().write_text(f"{index+1}\t{row[0]}\t{row[1]}\t{row[2]}\t{row[3]}") for index, row in enumerate(zip(dot_brackets, sequence_constraints, df['gc_content'], df['mfe']))]

def split_rfam_anta(data_dir, dataset, in_extension='.interim', out_extension='.rna'):
    path = Path(data_dir, dataset, f"{str(dataset)}{in_extension}").resolve()
    print('Write single files')
    with open(path) as f:
        for l in f:
            id = l.split()[0].rstrip()
            omega = l.split()[1].rstrip()
            psi = l.split()[2].rstrip().upper().replace('T', 'U')
            gc = id
            mfe = id
            Path(data_dir, dataset, f"{id}{out_extension}").resolve().write_text(f"{id}\t{omega}\t{psi}\t{gc}\t{mfe}")

def generate_sc_data(data_dir, dataset, in_extension='.interim', out_extension='.rna'):
    path = Path(data_dir, dataset, f"{str(dataset).split('_')[-1]}{in_extension}").resolve()
    print('Read interim data')
    df = pd.read_csv(path, sep='\t')
    print('Write single files')
    df['local_random'] = df['local_random'].apply(lambda x: x.upper().replace(')', 'N').replace('(', 'N').replace('.', 'N'))
    # print(df)
    # print(df['local_random'])
    out_path = Path(data_dir, dataset + '_sc')
    out_path.mkdir(exist_ok=True, parents=True)
    [Path(out_path,  f"{index+1}{out_extension}").resolve().write_text(f"{index+1}\t{row[0]}\t{row[1]}\t{row[2]}\t{row[3]}") for index, row in enumerate(zip(df['structure'], df['local_random'], df['gc_content'], df['mfe']))]

def generate_inverse_folding_data(data_dir, dataset, in_extension='.interim', out_extension='.rna'):
    path = Path(data_dir, dataset, f"{str(dataset).split('_')[-1]}{in_extension}").resolve()
    print('Read interim data')
    df = pd.read_csv(path, sep='\t')
    print('Write single files')
    df['sequence'] = df['sequence'].apply(lambda x: x.upper().replace('A', 'N').replace('G', 'N').replace('C', 'N').replace('U', 'N'))
    # print(df)
    print(df['sequence'])
    out_path = Path(data_dir, dataset + '_if')
    out_path.mkdir(exist_ok=True, parents=True)
    [Path(out_path,  f"{index+1}{out_extension}").resolve().write_text(f"{index+1}\t{row[0]}\t{row[1]}\t{row[2]}\t{row[3]}") for index, row in enumerate(zip(df['structure'], df['sequence'], df['gc_content'], df['mfe']))]

def generate_if_baseline_data(data_dir, dataset, in_extension='.interim', out_extension='.rna'):
    path = Path(data_dir, dataset, f"{str(dataset).split('_')[-1]}{in_extension}").resolve()
    print('Read interim data')
    df = pd.read_csv(path, sep='\t')
    print('Write single files')
    out_path = Path(data_dir, dataset + '_bl')
    out_path.mkdir(exist_ok=True, parents=True)
    [Path(out_path,  f"{index+1}{out_extension}").resolve().write_text(f"{row}") for index, row in enumerate(df['structure'])]

def generate_data_4_anta_sc_run(data_dir, dataset, in_extension='.interim', out_extension='.anta'):
    path = Path(data_dir, dataset, f"{str(dataset)}{in_extension}").resolve()
    print('Write single files')
    with open(path) as f:
        for l in f:
            id = l.split()[0].rstrip()
            omega = l.split()[1].rstrip()
            psi = l.split()[2].rstrip().upper().replace('T', 'U')
            gc = id
            mfe = id
            out_path = Path(data_dir, dataset + '_anta')
            out_path.mkdir(exist_ok=True, parents=True)
            Path(out_path, f"{id}{out_extension}").resolve().write_text(f"{omega}\t{psi}")

def generate_gap_data(data_dir, dataset, in_extension='.interim', out_extension='.rna'):
    path = Path(data_dir, dataset, f"{str(dataset).split('_')[-1]}{in_extension}").resolve()
    print('Read interim data')
    df = pd.read_csv(path, sep='\t')
    print('Introducing gaps')
    targets = []
    for tau in df['local_random']:
        tau_list = list(tau)
        iterations = np.random.randint(1, 10)
        current_index = 0
        for iteration in range(iterations):
            try:
                length = np.random.randint(1, int(0.1 * len(tau[current_index:])))
                start = np.random.randint(current_index + 1, len(tau) - length)
                tau_list[start:start + length] = 'N' * length
                current_index = start + 1
            except:
                break
        assert len(''.join(tau_list)) == len(tau)
        targets.append(''.join(tau_list))

    df['local_random'] = pd.DataFrame(targets, columns=['local_random'])

    out_dir = Path(data_dir, dataset + '_gaps')
    out_dir.mkdir(exist_ok=True, parents=True)

    print('Write single files')
    [Path(out_dir, f"{index+1}{out_extension}").resolve().write_text(f"{index+1}\t{row[0]}\t{row[1]}\t{row[2]}\t{row[3]}\t{row[4]}\t{row[5]}") for index, row in enumerate(zip(df['structure'], df['sequence'], df['local_random'], df['local_motif'], df['gc_content'], df['mfe']))]




if __name__ == '__main__':
    # split_local_data_to_single_files('data/rfam_learn_local/test')
    # split_local_data_to_single_files('data/rfam_learn_local/train')
    # split_local_data_to_single_files('data/rfam_learn_local/validation')
    # print("split test data")
    # split_via_dataframe('data/', 'rfam_local_short_train')
    # print('split training data')
    # split_via_dataframe('data/', 'rfam_local_long_train')
    # print('split validation data')
    # split_via_dataframe('data/rfam_learn_local', 'validation')
    # split_via_dataframe('data', 'eterna_local_test')
    # split_via_dataframe('data', 'rfam_taneda_local_test')
    # split_via_dataframe('data', 'rfam_local_test')
    # split_via_dataframe('data', 'rfam_local_min_400_max_1000_test')
    # split_via_dataframe('data', 'rfam_local_min_1000_test')
    # split_rfam_anta('data', 'rfam_anta_sc')
    # generate_sc_data('data', 'rfam_local_min_1000_test')
    # generate_inverse_folding_data('data', 'rfam_local_min_1000_test')
    # generate_if_baseline_data('data', 'rfam_local_min_1000_test')
    # generate_data_4_anta_sc_run('data', 'rfam_anta_sc')
	generate_gap_data('data', 'rfam_local_min_1000_test')



import numpy as np


def levenshtein_with_n(designed_sequence=None, designed_structure=None, target_sequence=None, target_structure=None):
    """
    Levenshtein distance between target_structure and designed_structure with 'N' for any char
    """
    # Initialisation for the table of calculations
    rows = len(target_structure) + 1
    cols = len(designed_structure) + 1
    table = [[0 for x in range(cols)] for x in range(rows)]

    for i in range(cols):
        table[0][i] = i

    for i in range(rows):
        table[i][0] = i

    # Calculation of the distances in the table
    for col in range(1, cols):
        for row in range(1, rows):
            # if same char cost for substitution is 0, else 1
            if target_structure[row - 1] == designed_structure[col - 1] or target_structure[row - 1] == 'N' or designed_structure[col - 1] == 'N':
                cost = 0
            else:
                cost = 1
            table[row][col] = min(table[row - 1][col] + 1,        # delete
                                  table[row][col - 1] + 1,        # insert
                                  table[row - 1][col - 1] + cost) # substitution

    #bottom right in the table is the levenshtein distance
    return table[row][col] / len(designed_structure)


def number_of_brackets_with_n(designed_sequence=None, designed_structure=None, target_sequence=None, target_structure=None):
    """
    difference of number of brackets between designed_structure and target_structure
    """
    mismatches =  abs(designed_structure.count('(') - target_structure.count('(')) + \
           abs(designed_structure.count(')') - target_structure.count(')')) + \
           abs(designed_structure.count('.') - target_structure.count('.'))

    return (mismatches / 2) - target_structure.count('N')


def hamming_with_n(designed_sequence=None, designed_structure=None, target_sequence=None, target_structure=None):
    l1 = np.asarray(list(designed_structure))
    l2 = np.asarray(list(target_structure))
    distance = np.sum(((l1 != l2) & (l1 != 'N') & (l2 != 'N')).astype(np.int8))
    return distance   #  / len(designed_structure)



import pandas as pd
from pathlib import Path


def parse_dot_brackets(
    dataset, data_dir, target_structure_ids=None, target_structure_path=None
):
    """TODO
    Generate the targets for next epoch.

    Args:
        dataset: The name of the benchmark to use targets from.
        data_dir: The directory of the target structures.
        target_structure_ids: Use specific targets by ids.
        target path: Specify a path to the targets.

    Returns:
        An epoch generator for the specified target structures.
    """
    if target_structure_path:
        target_paths = [target_structure_path]
    elif target_structure_ids:
        target_paths = [
            Path(data_dir, dataset, f"{id_}.rna") for id_ in target_structure_ids
        ]
    else:
        target_paths = list(Path(data_dir, dataset).glob("*.rna"))

    return [data_path.read_text().rstrip() for data_path in target_paths]

def parse_local_design_data(
    dataset, data_dir, target_structure_ids=None, target_structure_path=None
):
    """TODO
    Generate targets for next local design epoch.
    """
    if target_structure_path:
        return [tuple(Path(target_structure_path).read_text().rstrip().split())]
    # path = Path
    # df = pd.read_csv(Path(data_dir, dataset, f"{str(dataset).split('_')[-1]}.interim"), sep='\t')
    if target_structure_ids:
        rna_files = []
        for id in target_structure_ids:
            data = Path(data_dir, dataset, f"{id}.rna").read_text().split('\t')
            rna_files.append((data[0], data[1], data[2], data[3], data[4]))  # (id, structure_constraints, sequence_constraints, gc-content, mfe)
        # return [(index+1, row[0], row[1], row[2], row[3], row[4], row[5]) for index, row in enumerate(zip(df.iloc[[int(id)-1 for id in target_structure_ids]]['structure'], df.iloc[[int(id)-1 for id in target_structure_ids]]['sequence'], df.iloc[[int(id)-1 for id in target_structure_ids]]['local_random'], df.iloc[[int(id)-1 for id in target_structure_ids]]['local_motif'], df.iloc[[int(id)-1 for id in target_structure_ids]]['gc_content'], df.iloc[[int(id)-1 for id in target_structure_ids]]['mfe']))]
        return rna_files
    else:
        rna_files = []
        for path in Path(data_dir, dataset).glob('*.rna'):
            data = path.read_text().split('\t')
            rna_files.append((data[0], data[1], data[2], data[3], data[4]))
        return rna_files
        # return list(map(tuple, df[['structure', 'sequence', 'local_random', 'local_motif', 'gc_content', 'mfe']].itertuples(index=False, name=None)))
        # return [(index+1, row[0], row[1], row[2], row[3], row[4], row[5]) for index, row in enumerate(zip(df['structure'], df['sequence'], df['local_random'], df['local_motif'], df['gc_content'], df['mfe']))]



from tqdm import tqdm
import requests


def _download_dataset_from_http(url, download_path):
    response = requests.get(url, stream=True)
    with open(download_path, "wb+") as dataset_file:
        progress_bar = tqdm(
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            total=int(response.headers["Content-Length"]),
        )
        for data in tqdm(response.iter_content()):
            progress_bar.update(len(data))
            dataset_file.write(data)


def download_eterna(download_path):
    eterna_url = "https://ars.els-cdn.com/content/image/1-s2.0-S0022283615006567-mmc5.txt"
    _download_dataset_from_http(eterna_url, download_path)


def extract_secondarys(download_path, dump_path):
    with open(download_path) as input:
        parsed = list(zip(*(line.strip().split("\t") for line in input)))

    secondarys = parsed[4][1:]

    with open(dump_path, "w") as data_file:
        for structure in secondarys:
            if structure[-1] == "_":  # Weird dataset bug
                structure = structure[:-1]
            data_file.write(f"{structure}\n")


if __name__ == "__main__":
    download_path = f"data/eterna/raw/eterna_raw.txt"
    dump_path = f"data/eterna/interim/eterna.txt"
    download_eterna(download_path)
    extract_secondarys(download_path, dump_path)



from pathlib import Path

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data_path",
        default="data/rfam_local_test",
        help="Path to data directory",
    )
    parser.add_argument(
        "--dataset",
        default="rfam_local_test",
        help="Path to data directory",
    )

    args = parser.parse_args()

    dataset_path = Path(args.data_path, args.dataset)
    # out_dir = 'data/rfam_local_test_anta/'
    # out_dir_path = Path(out_dir)
    # out_dir_path.mkdir(exist_ok=True, parents=True)
    for path in dataset_path.glob('*.rna'):
        local = path.read_text().split('\t')[3]
        struc = path.read_text().split('\t')[1]
        # print(struc)
        seq = local.replace('.', 'N').replace('(', 'N').replace(')', 'N')
        out_path = Path(dataset_path, f"{path.stem}.anta")
        out_path.write_text(f"{struc}\t{seq}")



import os

os.environ["OMP_NUM_THREADS"] = "1"

import logging

logging.basicConfig(level=logging.DEBUG)

import argparse
import pickle
import time

import hpbandster.core.nameserver as hpns
import hpbandster.core.result as hpres

from hpbandster.optimizers import BOHB as BOHB
from src.optimization.learna_worker import LearnaWorker
from src.optimization.meta_learna_worker import MetaLearnaWorker


parser = argparse.ArgumentParser(
    description="Optimization pipeline for Learning to Design RNA"
)
parser.add_argument(
    "--min_budget",
    type=float,
    help="Minimum budget used during the optimization.",
    default=10,
)
parser.add_argument(
    "--max_budget",
    type=float,
    help="Maximum budget used during the optimization.",
    default=10,
)
parser.add_argument(
    "--n_iterations",
    type=int,
    help="Number of iterations performed by the optimizer",
    default=16,
)

parser.add_argument(
    "--n_cores", type=int, help="Number of workers to run in parallel.", default=10
)
parser.add_argument(
    "--worker", help="Flag to turn this into a worker process", action="store_true"
)

parser.add_argument(
    "--run_id",
    type=str,
    help="A unique run id for this optimization run."
    "An easy option is to use the job id of the clusters scheduler.",
)
parser.add_argument("--data_dir", type=str, help="path where the datasets are stored")
parser.add_argument(
    "--nic_name", type=str, help="Which network interface to use for communication."
)
parser.add_argument(
    "--shared_directory",
    type=str,
    help="A directory that is accessible for all processes, e.g. a NFS share.",
)

parser.add_argument("--mode", choices=["learna", "meta_learna"], default="learna")


# args=parser.parse_args("--run_id test --nic_name lo --shared_directory /tmp --n_cores 4 --data_dir src/data --mode L2DesignRNA".split())
args = parser.parse_args()

os.makedirs(args.shared_directory, exist_ok=True)

if args.mode == "learna":
    worker_cls = LearnaWorker
    worker_args = dict(
        data_dir=args.data_dir, num_cores=args.n_cores, train_sequences=range(1, 100, 3)
    )

if args.mode == "meta_learna":
    worker_cls = MetaLearnaWorker
    worker_args = dict(
        data_dir=args.data_dir,
        num_cores=args.n_cores,
        train_sequences=range(1, 65000),
        validation_timeout=60,
    )


# Every process has to lookup the hostname
host = hpns.nic_name_to_host(args.nic_name)

if args.worker:
    time.sleep(5)  # short artificial delay to make sure the nameserver is already running
    w = worker_cls(**worker_args, run_id=args.run_id, host=host)
    w.load_nameserver_credentials(working_directory=args.shared_directory)
    w.run(background=False)
    exit(0)
result_logger = hpres.json_result_logger(args.shared_directory, overwrite=True)


# Start a nameserver:
# We now start the nameserver with the host name from above and a random open port
NS = hpns.NameServer(
    run_id=args.run_id, host=host, port=0, working_directory=args.shared_directory
)
ns_host, ns_port = NS.start()

# Most optimizers are so computationally inexpensive that we can affort to run a
# worker in parallel to it. Note that this one has to run in the background to
# not plock!
w = worker_cls(
    **worker_args,
    run_id=args.run_id,
    host=host,
    nameserver=ns_host,
    nameserver_port=ns_port
)
w.run(background=True)

print(worker_cls.get_configspace())

# Run an optimizer
# We now have to specify the host, and the nameserver information
bohb = BOHB(
    configspace=worker_cls.get_configspace(),
    run_id=args.run_id,
    host=host,
    nameserver=ns_host,
    nameserver_port=ns_port,
    min_budget=args.min_budget,
    max_budget=args.max_budget,
    result_logger=result_logger,
    ping_interval=600,
    working_directory=args.shared_directory,
)
res = bohb.run(n_iterations=args.n_iterations, min_n_workers=20)


# In a cluster environment, you usually want to store the results for later analysis.
# One option is to simply pickle the Result object
with open(os.path.join(args.shared_directory, "results.pkl"), "wb") as fh:
    pickle.dump(res, fh)


# Step 4: Shutdown
# After the optimizer run, we must shutdown the master and the nameserver.
bohb.shutdown(shutdown_workers=True)
NS.shutdown()



#create a setup.py file to install the package
from setuptools import setup, find_packages


setup(
    name="learna_tools",
    version="0.1.0",
    description="RNA Design with automated reinforcement learning.",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[
        "Distance",
        "GraKeL==0.1.8",
        "numpy",
        "pandas",
        "pyaml",
        "pynisher",
        "pytest",
        "pylint",
        "psutil",
        "configspace",
        "scipy",
        "Sphinx",
        "tensorflow==1.4.0",
        "scikits.bootstrap",
        "hpbandster",
        "tqdm",
        "black",
        "tensorforce==0.3.3",
        "dataclasses",
        "munch",
        "pyyaml",
        "scikit-learn",
        "LinearFold",
        "tabulate",
        "matplotlib",
        "biopython",
        "logomaker",
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        # Add other relevant classifiers.
    ],
    scripts=[
        'bin/learna',
        'bin/meta-learna',
        'bin/meta-learna-adapt',
        'bin/liblearna',
    ],
# Optional
    author='Frederic Runge',
    author_email='runget@cs.uni-freiburg.de',
    license='MIT',
    keywords='RNA Design',
    url='https://github.com/automl/learna_tools.git'
)
